import io
import os
import pickle
import time
import typing

import numpy as np
import pandas as pd

from .data_struc_kit import sum_list


def adjust_file_block_pos(
        f: io.TextIOWrapper,
        start_pos: int,
        end_pos: int,
        split_symbol: str = '\n',
        adjust_direct: str = 'forward'
):
    """
    adjust_direct: 'forward' or 'backward'

    """
    # TODO
    if start_pos != 0:
        f.seek(start_pos - 1)
        if f.read(1) != '\n':
            line = f.readline()
            start_pos = f.tell()
    f.seek(start_pos)
    while (start_pos <= end_pos):
        line = f.readline()
        start_pos = f.tell()


def split_file_block(
        file: typing.Union[str, io.TextIOWrapper],
        mode='rb',
        block_num: int = None,
        block_adjust_symbol=None,
):
    if block_num is None:
        block_num = os.cpu_count()
    if isinstance(file, io.TextIOWrapper):
        file = file.name

    with open(file, mode) as f:
        file_size = f.seek(0, 2)
        if file_size < block_num:
            raise ValueError(f'Input file size is smaller than block number: {block_num} blocks for file {file}')
        block_size = int(file_size / block_num)

        # TODO check \n or other defined symbol in loop and adjust position
        #  `If block_adjust_symbol is not None: adjust_file_block_pos...`
        pos_list = []
        start_pos = 0
        for i in range(block_num):
            if i == block_num - 1:
                end_pos = file_size - 1
                pos_list.append((start_pos, end_pos))
                break
            end_pos = start_pos + block_size - 1
            if end_pos >= file_size:
                end_pos = file_size - 1
            if start_pos >= file_size:
                break
            pos_list.append((start_pos, end_pos))
            start_pos = end_pos + 1

    return pos_list


def file_prefix_time(with_dash=False):
    curr_time = time.strftime('%Y%m%d', time.localtime())
    prefix = curr_time + '-' if with_dash else curr_time
    return prefix


def pd_read_csv_skip_row(file, comment=None, **kwargs):
    if os.stat(file).st_size == 0:
        raise ValueError("File is empty")
    with open(file, 'r') as f:
        pos = 0
        cur_line = f.readline()
        while cur_line.startswith(comment):
            pos = f.tell()
            cur_line = f.readline()
            f.seek(pos)
    return pd.read_csv(f, **kwargs)


def read_one_col_file(file, skiprows=None):
    with open(file, 'r') as f:
        one_col_list = [row.strip('\n') for row in f.readlines()]
        one_col_list = one_col_list[skiprows:] if skiprows is not None else one_col_list
        while '' in one_col_list:
            one_col_list.remove('')
    return one_col_list


def flatten_two_headers_file(
        file,
        header_num=2,
        sep=',',
        method=None
) -> pd.DataFrame:
    """

    method: stack headers or cross-insert or lower-first

    Headle file with two headers like

    Peptide_Order	Peptide	Peptide_Mass	Modifications	Proteins
        Spectrum_Order	Title	Charge	Precursor_Mass
    1	AAAAAAAAAAAAAAAAAA	2000	Carbamidomethyl[C](9)	PAK
        1	T1	3	1999
        2	T2	3	1999
        3	T3	3	1999
        4	T1	3	1999
        5	T5	3	1999
    2	CCCCCCCCCCCCCCC	3000	Carbamidomethyl[C](15)	PBK
        1	T2	3	2999
    3	DDDDDDDDDDDDDDDD	4000	null	PCK
        1	T3	3	3999
        2	T1	3	3999
        3	T2	3	3999

    """
    if isinstance(file, str):
        if len(file) < 300 and os.path.exists(file):
            with open(file, 'r') as f:
                file = f.readlines()
        else:
            file = file.split('\n')

    headers = [file[i].rstrip(f'\n{sep}').split(sep) for i in range(header_num)]
    headers_used_col_idx = [[idx for idx, value in enumerate(header) if value != ''] for header in headers]
    headers_used_col_num = [len(idx) for idx in headers_used_col_idx]

    if method is None or method == 'stack':
        flatten_text_used_col_idx = []
        for idx, num in enumerate([0, *headers_used_col_num][:-1]):
            flatten_text_used_col_idx.append(np.arange(num, num + headers_used_col_num[idx]))
    elif method == 'cross-insert':
        flatten_text_used_col_idx = []
    elif method == 'lower-first':
        flatten_text_used_col_idx = []
    else:
        raise

    flatten_header = sum_list([[value for value in header if value != ''] for header in headers])
    flatten_col_num = len(flatten_header)
    flatten_text = []

    header_level = 1
    consensus_text = ['' for i in range(flatten_col_num)]
    for row in file[header_num:]:
        row = row.rstrip(f'\n{sep}').split(sep)
        for idx, value in enumerate(row, 1):
            if value != '':
                header_level = idx
                break
        if header_level == 1:
            consensus_text = ['' for i in range(flatten_col_num)]

        for value_idx, raw_idx in enumerate(headers_used_col_idx[header_level - 1]):
            consensus_text[flatten_text_used_col_idx[header_level - 1][value_idx]] = row[raw_idx]

        if header_level == header_num:
            flatten_text.append(consensus_text.copy())

    return pd.DataFrame(flatten_text, columns=flatten_header)


def process_list_or_file(x):
    if isinstance(x, list) or isinstance(x, set):
        target_list = x
    else:
        if os.path.isfile(x):
            target_list = read_one_col_file(x)
        else:
            raise
    return target_list


def print_path_basename_in_dict(path_dict: dict):
    for name, path in path_dict.items():
        print(f'{name}: {os.path.basename(path)}')


def print_path_exist():
    try:
        print(check_path)
    except FileNotFoundError:
        ...


def check_path(
        path: str,
        name: str = None,
        shown_path_right_idx: typing.Union[None, int, list, tuple] = 1,
        show_all_after_idx: bool = True,
        raise_error: bool = False,
        verbose: bool = False
):
    if shown_path_right_idx is None:
        shown_filepath = path
    elif isinstance(shown_path_right_idx, int):
        if shown_path_right_idx < 0:
            shown_filepath = path
        elif shown_path_right_idx == 0:
            shown_filepath = os.path.basename(path)
        else:
            split_path = path.split(os.path.sep)
            shown_filepath = os.path.sep.join(split_path[-shown_path_right_idx:]) if show_all_after_idx else split_path[-shown_path_right_idx]
    elif isinstance(shown_path_right_idx, (list, tuple)):
        # TODO join selected idx
        for idx in shown_path_right_idx:
            pass
    else:
        raise ValueError(f'Param `shown_path_right_idx` must be None or integer or list/tuple of integet. Now {shown_path_right_idx}')
    if name is not None:
        print(f'{os.path.exists(path)} - {name}: {shown_filepath}')
    else:
        print(f'{os.path.exists(path)} - {shown_filepath}')


def check_path_in_dict(path_dict: dict, shown_filename_right_idx: int = None):
    # TODO 显示的文件名称可以是多个 idx 对应 substring 的组合
    """
    :param path_dict:
    :param shown_filename_right_idx: None or int. None: use full path (raw value in dict). int: use idx part of file path (right count with 1st-first idx)
    """
    print(f'Total {len(path_dict)} files')
    for name, path in path_dict.items():
        if shown_filename_right_idx is None or (isinstance(shown_filename_right_idx, int) and shown_filename_right_idx == 0):
            shown_filename = path
        elif isinstance(shown_filename_right_idx, int) and shown_filename_right_idx > 0:
            used_path = path
            for _ in range(shown_filename_right_idx):
                used_path = os.path.dirname(used_path)
            shown_filename = os.path.basename(used_path)
        else:
            raise ValueError(f'Param `shown_filename_right_idx` must be None or any positive integer. Now {shown_filename_right_idx}')
        print(f'{os.path.exists(path)} - {name}: {os.path.basename(shown_filename)}')


def check_input_df(data, *args) -> pd.DataFrame:
    if isinstance(data, pd.DataFrame):
        df = data
    else:
        if os.path.exists(data):
            df = pd.read_csv(data, *args)
        else:
            raise FileNotFoundError
    return df


def fill_path_dict(path_to_fill: str, fill_string: dict, exist_path_dict: dict = None):
    if exist_path_dict is None:
        path_dict = dict()
    else:
        path_dict = exist_path_dict.copy()

    for k, file_name in fill_string.items():
        file_name = [file_name] if isinstance(file_name, str) else file_name
        path_dict[k] = path_to_fill.format(*file_name)
    return path_dict


def join_path(path, *paths, create=False):
    pass


def write_inten_to_json(prec_inten: dict, file_path):
    total_prec = len(prec_inten)
    with open(file_path, 'w') as f:
        f.write('{\n')

        for prec_idx, (prec, inten_dict) in enumerate(prec_inten.items(), 1):
            f.write('    "%s": {\n' % prec)
            frag_num = len(inten_dict)
            for frag_idx, (frag, i) in enumerate(inten_dict.items(), 1):
                if frag_idx != frag_num:
                    f.write(f'        "{frag}": {i},\n')
                else:
                    f.write(f'        "{frag}": {i}\n')

            if prec_idx != total_prec:
                f.write('    },\n')
            else:
                f.write('    }\n')

        f.write('}')


def data_dump_load_skip(file_path, data=None, cover_data=False, update_file=False):
    if not os.path.exists(file_path):
        if data is not None:  # Here use 'is not None' because some thing will be wrong when the data is a pd.DataFrame. (Truth value is ambiguous error)
            with open(file_path, 'wb') as f:
                pickle.dump(data, f)
        else:
            raise FileNotFoundError('No existing file and no input data')
    else:
        if data is not None:
            if update_file:
                with open(file_path, 'wb') as f:
                    pickle.dump(data, f)
            elif cover_data:
                with open(file_path, 'rb') as f:
                    data = pickle.load(f)
            else:
                pass
        else:
            with open(file_path, 'rb') as f:
                data = pickle.load(f)
    return data



import os

import numpy as np
import pandas as pd
from tqdm import tqdm

from mskit import rapid_kit


def processing_progress_logger(progress_type=None, print_progress=True, save_log=False):
    pass


class SpectronautResults(object):
    """
    This will store multi results as a dict
    """

    def __init__(self, result_path_dict=None, read_within_init=False, ident_func=lambda x: os.path.basename(x)):  # 这里可以是list然后传一个ident func来拿到result name或dict直接给定result name
        self.result_path_dict = result_path_dict
        self.filename2key_func = ident_func
        self.results = dict()

        self.total_result_file_num = 0

        if read_within_init:
            self.read_results()

        self.total_result_data = None
        self.phos_result_data = None

        self.stats_number_matrix = None
        self.stats_number_df = None

        self.exp_names = None
        self.run_names = None

    def __len__(self):
        return len(self.results)

    def __add__(self, other: dict):
        pass

    def add(self, other):
        self.__add__(other)

    def empty_cache(self):
        """
        Remove all additional information and stored data except the raw input data and parameters
        """

    def read_results(self, result_path_dict=None, print_progress=False):
        if result_path_dict:
            self.result_path_dict = result_path_dict
        self.total_result_file_num = len(self.result_path_dict)
        for i, (name, f) in enumerate(self.result_path_dict.items()):
            if print_progress:
                print(f'Loading result file {i + 1}/{self.total_result_file_num}')
            self.results[name] = SpectronautResult(f)

    def remove_all_results(self):
        self.results = dict()

    def add_predefined_features_to_all(self, ptm_thres=0.99, print_progress=False):
        for name, result_df in self.results.items():
            result_df.add_predefined_features(ptm_thres=ptm_thres)

    def get_results_stats(self):
        self.total_result_data = dict()
        self.phos_result_data = dict()
        for col in ['FirstProtein', 'Prec', 'EG.ModifiedPeptide', 'ProtPhosPosComb']:
            self.total_result_data[col] = dict()
            self.phos_result_data[col] = dict()
            for lib_type, sn_result in self.results.items():
                result_df = sn_result.return_result_df()
                if col == 'ProtPhosPosComb':
                    self.total_result_data[col][lib_type] = set(';'.join(result_df[col].dropna().tolist()).split(';'))
                    self.phos_result_data[col][lib_type] = set(';'.join(result_df[result_df['IsPhospep']][col].dropna().tolist()).split(';'))
                else:
                    self.total_result_data[col][lib_type] = set(result_df[col])
                    self.phos_result_data[col][lib_type] = set(result_df[result_df['IsPhospep']][col])

    def construct_result_stats_matrix(self):
        self.stats_number_matrix = np.array([
            sum([
                [
                    len(self.total_result_data[col][lib_type]),
                    len(self.phos_result_data[col][lib_type])
                ] for col in ['FirstProtein', 'Prec', 'EG.ModifiedPeptide', 'ProtPhosPosComb']
            ], []) for lib_type in self.results.keys()
        ])

    def construct_result_stats_df(self):
        self.stats_number_df = pd.DataFrame(
            self.stats_number_matrix[:, :-1],
            index=list(self.results.keys()),
            columns=sum([[_, _ + '-PhosOnly'] for _ in ['Protein', 'Prec', 'Modpep', 'Phossite']], [])[:-1],
        )

    def return_result_dfs(self):
        return self.results


class SpectronautResult(object):
    def __init__(self, result_file=None):

        # The path of result file and read the result file if the file path is passed when news a object of this class
        self.result_file = result_file
        if self.result_file:
            self.read_result(self.result_file)
        else:
            self.result_df = None

        self.initial_cols = None

        self.decoy_on = False  # if IsDecoy in the columns and True in the values -> decoy_on = True

    def read_result(self, result_file):
        """
        :param result_file The path of result file from SN (Normal scheme)
        This func can be called any time to set a result file.
        And the func init_results will also be called at first, then the previously stored varibles will be removed.
        """
        self.init_results()
        if isinstance(result_file, pd.DataFrame):
            self.result_df = result_file
        else:
            self.result_file = result_file
            self.result_df = pd.read_csv(result_file, sep='\t', low_memory=False)
        self.initial_cols = self.result_df.columns

    def init_results(self):
        """
        This will set all stored contants and results to None
        """
        self.result_file = None
        self.result_df = None
        self.initial_cols = None

    def add_first_protein(self):
        self.result_df['FirstProtein'] = self.result_df['PG.ProteinGroups'].apply(lambda x: x.split(';')[0])

    def add_prec(self):
        self.result_df['Prec'] = self.result_df['EG.ModifiedPeptide'] + '.' + self.result_df['FG.Charge'].astype(int).astype(str)

    def add_intseq(self):
        str_mod_to_int_dict = {
            'C[Carbamidomethyl (C)]': 'C',
            'M[Oxidation (M)]': '1',
            'S[Phospho (STY)]': '2',
            'T[Phospho (STY)]': '3',
            'Y[Phospho (STY)]': '4',
        }

        def trans_str_mod_to_int(pep):
            pep = pep.replace('_', '')
            for mod, int_mod in str_mod_to_int_dict.items():
                pep = pep.replace(mod, int_mod)
            return pep

        self.result_df['IntSeq'] = self.result_df['EG.ModifiedPeptide'].apply(trans_str_mod_to_int)

    def add_is_phos(self):
        self.result_df['IsPhospep'] = self.result_df['EG.PTMPositions [Phospho (STY)]'].apply(lambda x: True if pd.notna(x) else False)

    def add_phossite_in_pep(self, ptm_thres=0.99):
        def _add_pep_site(x, p_t):
            if not x['IsPhospep']:
                return np.nan
            prob_list = x['EG.PTMProbabilities [Phospho (STY)]'].split(';')
            site = [s for i, s in enumerate(x['EG.PTMPositions [Phospho (STY)]'].split(';')) if float(prob_list[i]) >= p_t]
            return ';'.join(site)

        self.result_df['PepPhosPos'] = self.result_df.apply(_add_pep_site, args=(ptm_thres,), axis=1)

    def add_phossite_in_prot(self):
        def _add_prot_site(x):
            if pd.isna(x['PepPhosPos']):
                return np.nan
            pep_pos = int(x['PEP.PeptidePosition'].split(';')[0].split(',')[0])
            phos_site_in_prot = [int(_) + pep_pos - 1 for _ in x['PepPhosPos'].split(';')]
            return ';'.join(map(str, phos_site_in_prot))

        self.result_df['ProtPhosPos'] = self.result_df.apply(_add_prot_site, axis=1)

    def add_protein_phossite_comb(self):
        def _add_comb_site(x):
            if pd.isna(x['ProtPhosPos']):
                return np.nan
            else:
                return ';'.join([x['FirstProtein'] + ',' + _ for _ in x['ProtPhosPos'].split(';')])

        self.result_df['ProtPhosPosComb'] = self.result_df.apply(_add_comb_site, axis=1)

    def add_predefined_features(self, feature_list=None, **kwargs):
        """
        :param feature_list TODO 最后应该把所有feature单独添加，这里传一个list来选择添加哪些feature
        """
        self.add_first_protein()
        self.add_prec()
        self.add_intseq()
        self.add_is_phos()
        self.add_phossite_in_pep(ptm_thres=kwargs['ptm_thres'])
        self.add_phossite_in_prot()
        self.add_protein_phossite_comb()

    def count_loss_type(self):
        return self.result_df['F.FrgLossType'].value_counts()

    def __getitem__(self, item):
        return self.result_df[item]

    def __setitem__(self, key, value):
        self.result_df[key] = value

    def get_columns(self):
        return self.result_df.columns

    def set_columns(self, cols):
        if not isinstance(cols, list):
            cols = [cols]
        self.result_df.columns = cols

    columns = property(get_columns, set_columns, doc='''Columns of result df''')

    def get_frag_inten(self) -> dict:
        result_spec = dict()
        reps = self.result_df['R.Replicate'].drop_duplicates().tolist()  # 这里应该增加可以选择 Replicate 或是 FileName
        with tqdm(reps) as t:
            for rep in t:
                one_rep_result = self.result_df[self.result_df['R.Replicate'] == rep]
                one_rep_spec_dict = dict()
                for prec, df in one_rep_result.groupby('FG.Id'):  # TODO FG.Id 不在 result 列中
                    frag_dict = dict()
                    for row_index, row in df.iterrows():
                        frag_dict[f'{row["F.FrgIon"]}+{row["F.Charge"]}-{row["F.FrgLossType"]}'] = row['F.NormalizedPeakArea']
                    one_rep_spec_dict[prec] = frag_dict
                result_spec[rep] = one_rep_spec_dict
        return result_spec

    def recalc_exp_irt(self):
        self.result_df['ReCalciRT'] = self.result_df.apply(
            lambda x: (x['EG.MeanApexRT'] - x['EG.StartRT']) / (
                    x['EG.EndRT'] - x['EG.StartRT']) * (
                              x['EG.EndiRT'] - x['EG.StartiRT']) + x['EG.StartiRT'], axis=1)

    def return_result_df(self, original_cols=False):
        return self.result_df


class SpectronautResultPivotPeptide(SpectronautResult):
    """
    TODO
    1. 读取search result
    2. 对search result加入protein list分类
    3. 提取相同的rep（给定ident分为对应的result）
    """

    def __init__(self):
        super(SpectronautResultPivotPeptide, self).__init__()


class SpectronautResultPivotProtein(SpectronautResult):
    def __init__(self):
        super(SpectronautResultPivotProtein, self).__init__()


# TODO: Re-write this


def get_one_prefix_result(result_df, prefix, suffixes):
    return [set(result_df[f'{prefix}-{each_suffix}'].dropna()) for each_suffix in suffixes]


def get_search_result(result_file, sheet_names, prefixes=('Protein', 'Precursor'), suffixes=('R1', 'R2', 'R3')):
    if isinstance(sheet_names, str):
        sheet_names = [sheet_names]
    with pd.ExcelFile(result_file) as f:
        result_list = []
        for sheet in sheet_names:
            df = f.parse(sheet_name=sheet)
            store_dict = dict()
            for each_prefix in prefixes:
                store_dict[each_prefix] = get_one_prefix_result(df, each_prefix, suffixes)
                store_dict[f'{each_prefix}-Total'] = rapid_kit.data_struc_kit.sum_set_in_list(
                    store_dict[each_prefix])
            if 'Precursor' in prefixes:
                store_dict['Peptide'] = [set([_.split(
                    '.')[0] for _ in each_suff_data]) for each_suff_data in store_dict[each_prefix]]
                store_dict['Peptide-Total'] = rapid_kit.data_struc_kit.sum_set_in_list(
                    store_dict['Peptide'])
            result_list.append(store_dict)
    return result_list


def select_target_df(original_df, region_identifier):
    region_df = rapid_kit.extract_df_with_col_ident(original_df, region_identifier, focus_col='R.Instrument (parsed from filename)')
    return region_df


def read_search_result_intensity(result_file):
    result_df = pd.read_csv(result_file, sep='\t', low_memory=False)
    region_intensity_list = []
    for _ in ['region3', 'region5', 'region6']:
        region_intensity_dict = dict()
        each_region_df = select_target_df(result_df, _)
        for each_prec in each_region_df['EG.PrecursorId'].drop_duplicates():
            each_prec_df = each_region_df[each_region_df['EG.PrecursorId'] == each_prec]
            noloss_prec_df = each_prec_df[each_prec_df['F.FrgLossType'] == 'noloss']
            fragment_list = (noloss_prec_df['F.FrgType'] +
                             noloss_prec_df['F.FrgNum'].astype(str) + '+' +
                             noloss_prec_df['F.Charge'].astype(str)).tolist()
            fragment_intensity = noloss_prec_df['F.MeasuredRelativeIntensity'].tolist(
            )
            region_intensity_dict[each_prec] = dict(
                zip(fragment_list, fragment_intensity))
        region_intensity_list.append(region_intensity_dict)
    return region_intensity_list



from collections import defaultdict

import matplotlib.pyplot as plt
import pandas as pd

pFind = 3


class PFindConst:
    ProteinSummaryComment = '----Summary----'
    ProteinCutoffComment = '-' * 40


class ProteinFile(object):
    """
    REV_sp
    ...
    REV_CON
    """

    def __init__(self, file_path):
        self.file_path = file_path

        self.main_title = None
        self.sub_title = None

        self.protein_group_title = ['']

        self.protein_list = list()
        self.prec_list = list()
        self.prec_list = list()

    def _parse_protein_file(self):
        with open(self.file_path, 'r') as f:
            self.main_title = f.readline().strip('\n').split('\t')
            self.sub_title = f.readline().strip('\n\t').split('\t')
            for row in f:
                strip_row = row.strip('\n\t')
                if row == PFindConst.ProteinCutoffComment:
                    pass

    def pfind_protein_df(self, p, remove_rev=True, remove_con=False):
        protein_main = []
        protein_sub = []
        with open(p, 'r') as f:
            main_title = f.readline().strip('\n').split('\t')
            f.readline()
            for row in f:
                if '-' * 40 in row:
                    continue
                if '----Summary' in row:
                    break
                split_row = row.strip('\n').split('\t')
                if split_row[0].isdigit():
                    current_protein = split_row[1]
                    main_score = split_row[2]
                    main_q = split_row[3]
                    protein_main.append(split_row[:-1])
                else:
                    if split_row[1]:
                        protein_sub.append([current_protein, main_score, main_q, split_row[1], split_row[2], split_row[3]])
        protein_main_df = pd.DataFrame(protein_main, columns=main_title)
        protein_sub_df = pd.DataFrame(protein_sub, columns=['MainProtein', 'MainScore', 'MainQ-Value', 'Type', 'Protein', 'Q-Value'])
        if remove_rev:
            protein_main_df = protein_main_df[~protein_main_df['AC'].str.contains('REV_')]
            protein_sub_df = protein_sub_df[~protein_sub_df['Protein'].str.contains('REV_')]
        if remove_con:
            protein_main_df = protein_main_df[~protein_main_df['AC'].str.contains('CON_')]
            protein_sub_df = protein_sub_df[~protein_sub_df['Protein'].str.contains('CON_')]
        return protein_main_df, protein_sub_df

    def get_protein_list(self):
        Protein_MainDF, Protein_SubDF = self.pfind_protein_df(PATH_pFindResults['Test'], remove_rev=True, remove_con=True)
        ProtMain = set(Protein_MainDF[Protein_MainDF['Q-Value'].astype(float) < 0.01]['AC'])
        ProtMain_ACC = [_.split('|')[1] for _ in ProtMain]
        ProtTotal = set(Protein_MainDF['AC']) | set(Protein_SubDF['Protein'])
        ProtTotal_ACC = [_.split('|')[1] for _ in ProtTotal]
        print(len(ProtMain_ACC))
        print(len(ProtTotal_ACC))

    def plot_protein_num(self):
        plt.figure(figsize=(2.5, 1.), dpi=300)
        for i, protein_list in enumerate([ProtMain_ACC, ProtTotal_ACC]):
            plt.bar(i, len(protein_list), )
            plt.annotate(format(len(protein_list), ','), (i, len(protein_list) / 2), va='center', ha='center')
        bpk.drawing_area.set_thousand_separate()
        plt.xticks(list(range(2)), ['Main', 'Total'], rotation=45)
        plt.title('Total proteins in results')


class SpecFile(object):
    def __init__(self):
        pass

    @staticmethod
    def pfind_spec_prot_to_file(x):  # 把这个函数放在外面
        proteins = x['Proteins'].strip('/').split('/')
        file = x['File_Name'].split('.')[0]
        prot_len = len(proteins)
        return dict(zip(proteins, [file] * prot_len))

    def get_prot_to_file_dict(self):
        prot_to_file_list = SpecDF.apply(self.pfind_spec_prot_to_file, axis=1).tolist()
        prot_to_file = defaultdict(set)
        for pro_file in prot_to_file_list:
            for name, file in pro_file.items():
                if name.startswith('REV_') or name.startswith('CON_'):
                    continue
                for r_index, region_ident in enumerate(RegionIdent):
                    if region_ident in file:
                        prot_to_file[name.split('|')[1]].add(f'R{r_index + 1}')
        print('Total prot -> file dict (contains REV_):', len(prot_to_file.keys()))
        return prot_to_file



import os
import re
from collections import Counter

import pandas as pd
import prettytable

from .sn_constant import *
from .sn_utils import *

try:
    from sklearn.model_selection import train_test_split
except ModuleNotFoundError:
    pass

from mskit import rapid_kit
from mskit.calc import normalize_intensity


class SpectronautLibrarys(object):
    def __init__(self, lib_dict, init_oper=None):
        pass


class SpectronautLibrary(object):
    """
    self._library_storage is used for the original library dataframe storage so that self._lib_df can be restored when it is filtered.

    限定每列的 type 为 str，读取之后再转换指定列的格式
    """
    # TODO 提 fragment intensity 这种操作可以加并行，把提取步骤提出来，函数只负责做循环然后加一个装饰器？或是直接写两个方法

    LibBasicCols = [
        'PrecursorCharge',
        'ModifiedPeptide', 'StrippedPeptide',
        'iRT',
        'LabeledPeptide', 'PrecursorMz',
        'FragmentLossType', 'FragmentNumber', 'FragmentType', 'FragmentCharge', 'FragmentMz',
        'RelativeIntensity',
    ]
    FragInfoCols = [
        'FragmentLossType',
        'FragmentNumber',
        'FragmentType',
        'FragmentCharge',
        'FragmentMz',
        'RelativeIntensity',
    ]
    LibBasicCols_WithoutFrag = rapid_kit.subtract_list(LibBasicCols, FragInfoCols)
    LibBasicCols_WithPG = [*LibBasicCols, 'ProteinGroups']

    def __init__(self, lib=None, spectronaut_version=14, pd_low_mem=False):
        self._spectronaut_version = spectronaut_version
        self._Mod = ModType(self._spectronaut_version)

        self._lib_path = None
        self._lib_df = None
        self._library_storage = None
        self._initial_lib_title = None

        self._lib_name = None

        self._prec = None
        self._modpep = None
        self._stripped_pep = None

        if lib is not None:
            self.set_library(lib, pd_low_mem=pd_low_mem)
            if isinstance(lib, str):
                self._lib_path = lib

    def init(self):
        pass

    def partial_init(self):
        pass

    def full_init(self):
        pass

    def require(self, required_list):
        require_dict = {
            'Prec': [self.add_prec()],
            'IntPep': [self.add_intpep()],
            'IntPrec': [self.add_prec(), self.add_intprec()]
        }
        return [require_dict[k] for k in required_list]

    def set_library(self, lib, pd_low_mem=False):
        if isinstance(lib, pd.DataFrame):
            self._lib_df = lib
        else:
            if os.path.isfile(lib):
                self._lib_path = lib
                self._lib_df = pd.read_csv(self._lib_path, sep='\t', low_memory=pd_low_mem)
            else:
                raise FileNotFoundError(f'The input library is not a dataframe and not a valid path.\n'
                                        f'Now is {type(lib)}: {lib}')

        self._initial_lib_title = self._lib_df.columns
        self._library_storage = self._lib_df.copy()

    def save(self, path, sep='\t', create_folder_if_notexist=False, save_func=None):
        """
        self.write_current_library
        """
        self._lib_df.to_csv(path, sep=sep, index=False)

    def __len__(self):
        return len(self._lib_df)

    def __add__(self, other):
        """
        Directly add the two libraries with no further operation
        """
        pass

    def __iadd__(self, other):
        pass

    def __mul__(self, other):
        """
        Let the fragments in two libraries combined together
        """
        pass

    def __getitem__(self, item):
        return self._lib_df[item]

    def __setitem__(self, key, value):
        self._lib_df[key] = value

    def columns(self):
        pass

    def fragment_loss_counts(self, col='FragmentLossType', readable_trans=True):
        loss_count = dict(Counter(self._lib_df[col]))
        if readable_trans:
            temp = dict()
            for loss_name, num in loss_count.items():
                if loss_name in LossType.SN_to_Readable:
                    temp[LossType.SN_to_Readable[loss_name]] = num
                else:
                    temp[loss_name] = num
            loss_count = temp.copy()
            del temp
        return loss_count

    def fragment_number_counts(self):
        pass

    def fragment_charge_counts(self):
        pass

    def summary(self, sum_list=None):
        summary_funcs = {
            'FragmentLossType': self.fragment_loss_counts,
        }
        summary_tabel_title = {
            'FragmentLossType': ['LossType', 'Number'],
        }
        long_sep = '-' * 20
        short_sep = '-' * 5
        if sum_list is None:
            sum_list = ['FragmentLossType', ]

        summary_contents = ''

        for each_feature in sum_list:
            sum_string = ''

            used_func = summary_funcs[each_feature]
            results = used_func()
            write_table = prettytable.PrettyTable(field_names=summary_tabel_title[each_feature])
            for each_result_name, result_num in results.items():
                write_table.add_row([each_result_name, result_num])

            sum_string += f'{long_sep}\n'
            sum_string += f'{short_sep} {each_feature} {short_sep}\n'
            sum_string += f'{write_table.get_string()}\n'
            sum_string += f'{long_sep}\n'
            print(sum_string)
            summary_contents += '\n'
            summary_contents += sum_string

    def backtrack_library(self):
        """
        replace lib_df with storage
        """
        self._lib_df = self._library_storage

    def store_current_library(self):
        self._library_storage = self._lib_df.copy()

    def switch_curr_store_lib(self):
        _ = self._lib_df.copy()
        self._lib_df = self._library_storage.copy()
        self._library_storage = _

    def add_prec(self):
        if 'Precursor' not in self._lib_df:
            self._lib_df['Precursor'] = self._lib_df['ModifiedPeptide'] + '.' + self._lib_df['PrecursorCharge'].astype(str)

    def add_intpep(self):
        # TODO This should be trans pep format (need a general function)
        self._lib_df['IntPep'] = self._lib_df['ModifiedPeptide'].apply(sn_modpep_to_intseq)

    def add_intprec(self):
        self._lib_df['IntPrec'] = self._lib_df['IntPep'] + '.' + self._lib_df['PrecursorCharge'].astype(str)

    def add_first_protein(self, target_col='ProteinGroups', add_col='FirstProtein'):
        def split_protein(x, col):
            acc = x[col]
            if pd.isna(acc):
                return np.nan
            acc = acc.split(';')[0].strip()
            return acc

        self._lib_df[add_col] = self._lib_df.apply(split_protein, col=target_col, axis=1)

    def add_protein_type(self, type_dict, target_col='ProteinGroups', add_col='ProteinType', add_method=lambda x: x.split(';')[0], split_sign=';'):
        """
        lambda x: x for first protein
        lambda x: x.split(';')[0] for ProteinGroups
        'All' to map type dict to all proteins in the cell. This needs the param split_sign to define the sign for splitting proteins.
        """
        pass

    def assign_protein(self, seq_dict, add_col='AssignedProteins'):
        pass

    def assign_protein_with_ref(self, ref_dict, seq_dict=None,
                                key_level='StrippedPeptide', add_col='AssignedProteins'):
        if seq_dict is not None:
            pass

    def trans_frag_losstype(self, trans_dict):
        self._lib_df['FragmentLossType_Raw'] = self._lib_df['FragmentLossType']
        self._lib_df['FragmentLossType'] = self._lib_df['FragmentLossType_Raw'].replace(trans_dict)

    def add_frag_name(self, trans_dict=None, used_losstype_col='FragmentLossType'):
        if trans_dict is not None:
            self.trans_frag_losstype(trans_dict)
        self._lib_df['FragName'] = self._lib_df["FragmentType"] + \
                                   self._lib_df["FragmentNumber"].astype(str) + '+' + \
                                   self._lib_df["FragmentCharge"].astype(str) + '-' + \
                                   self._lib_df[used_losstype_col]

    def add_inten_rank(self, base_col='Precursor', inten_col='RelativeIntensity', **rank_args):
        if 'method' not in rank_args:
            rank_args['method'] = 'min'
        self._lib_df['IntenRank'] = self._lib_df.groupby(base_col)[inten_col].rank(**rank_args).astype(int)

    def retain_basic_cols(self):
        self._lib_df = self._lib_df[self.LibBasicCols]

    def add_predefined_features(self, feature_list=None):
        """
        :param feature_list TODO 最后应该把所有feature单独添加，这里传一个list来选择添加哪些feature
        """
        self.add_prec()
        self.add_intpep()
        self.add_intprec()

    def remove_mod(self, mods, regex=False):
        mods = [mods] if isinstance(mods, str) else mods
        for mod in mods:
            self._lib_df = self._lib_df[~self._lib_df['ModifiedPeptide'].str.contains(mod, regex=regex)]

    def get_all_mods(self, sort=True, modpep_col='ModifiedPeptide'):
        mods = []
        for one_pep_mod in [re.findall(r'\[.+?\]', _) for _ in list(set(self._lib_df[modpep_col]))]:
            for mod in one_pep_mod:
                if mod not in mods:
                    mods.append(mod)
        if sort:
            mods = sorted(mods)
        return mods

    def count_mods(self, level='single'):
        """
        "single" to count each mod in all peps, "all" to count mod combination
        """
        pass

    def count_frag_num(self, prec_col='IntPrec'):
        return self._lib_df.groupby(prec_col).apply(len).value_counts().sort_index()

    def get_all_aa(self, sort=True, used_col='StrippedPeptide'):
        aas = []
        for one_pep_aa in [list(_) for _ in list(set(self._lib_df[used_col]))]:
            for aa in one_pep_aa:
                if aa not in aas:
                    aas.append(aa)
        if sort:
            aas = sorted(aas)
        return aas

    def get_frag_inten(self, prec_col='Precursor') -> dict:
        lib_spec = dict()
        if 'FragName' not in self._lib_df.columns:
            self.add_frag_name()
        for prec, df in self._lib_df.groupby(prec_col):
            lib_spec[prec] = dict(df[['FragName', 'RelativeIntensity']].values)
        return lib_spec

    def get_rt_df(self, pep_col='ModifiedPeptide', rt_col='iRT') -> pd.DataFrame:
        rt_df = self._lib_df[[pep_col, rt_col]].drop_duplicates(pep_col)
        return rt_df

    def get_frag_inten_phos(self, prec_col='Precursor', output_key_col=None,
                            limit_charge=(1, 2),
                            limit_loss=('noloss', 'H3PO4'),
                            index_loss='H3PO4', min_frag_num=None) -> dict:
        """
        这个现在只用来作为 pDeep2 的 test 时提取 library 中的 Intensity
        如果是用于 training 提取 library 中的 fragment Intensity 使用 get_frag_inten
        之后应该把这个里面对 Phos loss 最小 by 离子位置的功能放在外面
        """
        lib_spec = dict()
        if 'FragName' not in self._lib_df.columns:
            self.add_frag_name()
        if output_key_col is None:
            output_key_col = prec_col
        if min_frag_num is None:
            min_frag_num = 0
        for prec, df in self._lib_df.groupby(prec_col):
            modpep, charge = prec.split('.')
            output_key = df.iloc[0][output_key_col]
            strip_pep, mod_sites, mods = rapid_kit.substring_finder(modpep.replace('_', ''))
            phos_mod_site = [mod_sites[i] for i, mod in enumerate(mods) if mod == '[Phospho (STY)]']
            st_phosmod_site = [s for s in phos_mod_site if strip_pep[s - 1] in ('S', 'T')]
            if st_phosmod_site:
                positive_min_phos_site = min(st_phosmod_site)
                negetive_min_phos_site = len(strip_pep) - max(st_phosmod_site) + 1
            else:
                positive_min_phos_site = len(strip_pep) + 1
                negetive_min_phos_site = len(strip_pep) + 1
            frag_dict = dict()
            for row_index, row in df.iterrows():
                frag_type = row["FragmentType"]
                frag_num = row["FragmentNumber"]
                frag_charge = row["FragmentCharge"]
                frag_losstype = row["FragmentLossType"]
                if frag_charge not in limit_charge:
                    continue
                if frag_losstype not in limit_loss:
                    continue
                if frag_losstype == index_loss:
                    if frag_type == 'b' and frag_num < positive_min_phos_site:
                        continue
                    elif frag_type == 'y' and frag_num < negetive_min_phos_site:
                        continue
                    else:
                        pass
                frag_dict[row['FragName']] = row['RelativeIntensity']
            if len(frag_dict) >= min_frag_num:
                lib_spec[output_key] = normalize_intensity(frag_dict, max_num=100)
        return lib_spec

    def remove_library_add_col(self):
        new_cols = [_ for _ in self._lib_df.columns if _ in self._initial_lib_title]
        self._lib_df = self._lib_df[new_cols]

    def return_library(self, remove_add_col=False):
        if remove_add_col:
            self.remove_library_add_col()
        return self._lib_df.copy()

    to_df = return_library

    def write_current_library(self, output_path, remove_add_col=True):
        """
        write current lib to file
        """
        if remove_add_col:
            self.remove_library_add_col()
        self._lib_df.to_csv(output_path, index=False, sep='\t')

    def split_library(self, split_size=(0.9, 0.1), split_by='ModifiedPeptide', save=None, seed=0):
        by_list = self._lib_df[split_by].drop_duplicates().tolist()
        if len(split_size) == 2:
            train_list, test_list = train_test_split(by_list, train_size=split_size[0], test_size=split_size[1], random_state=seed)
            if save is None:
                return (self._lib_df[self._lib_df[split_by].isin(train_list)],
                        self._lib_df[self._lib_df[split_by].isin(test_list)])
            else:
                pass
        elif len(split_size) == 3:
            size_outof_train = split_size[1] + split_size[2]
            train_list, val_test_list = train_test_split(by_list, train_size=split_size[0], test_size=size_outof_train, random_state=seed)
            val_size, test_size = split_size[1] / size_outof_train, split_size[2] / size_outof_train
            val_list, test_list = train_test_split(val_test_list, train_size=val_size, test_size=test_size, random_state=seed)
            return (self._lib_df[self._lib_df[split_by].isin(train_list)],
                    self._lib_df[self._lib_df[split_by].isin(val_list)],
                    self._lib_df[self._lib_df[split_by].isin(test_list)])
        else:
            pass

    def library_basic_info(self):
        """
        Count basic library infomation includes
            'protein_group', 'protein', 'modpep', 'prec', 'stripped_pep'
        :return:
        """
        protein_group = set(self._lib_df['ProteinGroups'])
        protein = set(sum([_.split(';') for _ in protein_group], []))
        self._modpep = set(self._lib_df['ModifiedPeptide'])
        self.add_prec()
        self._prec = set(self._lib_df['Precursor'])
        self._stripped_pep = set(self._lib_df['StrippedPeptide'])
        library_info_dict = dict((('ProteinGroup', len(protein_group)),
                                  ('SingleProtein', len(protein)),
                                  ('ModifiedPeptide', len(self._modpep)),
                                  ('Precursor', len(self._prec)),
                                  ('StrippedPeptide', len(self._stripped_pep)),
                                  ))
        charge = self._lib_df['PrecursorCharge'].tolist()
        return library_info_dict

    def strippep_length_distrib(self) -> pd.Series:
        """
        Count peptides length in the library
        :return: pandas.Series, which has all peptide length as index and corresponding peptide number as value. The index is sorted increasely.
        """
        stripped_pep = self._lib_df['StrippedPeptide'].drop_duplicates()
        stripped_pep_len = stripped_pep.str.len()
        len_ser = stripped_pep_len.value_counts()
        sorted_len_ser = len_ser.sort_index()
        return sorted_len_ser

    def strippep_miss_cleavage_distrib(self) -> pd.Series:
        """
        Count miss cleavage in the library
        :return: pandas.Series, which has miss cleavage number as index and corresponding peptide number as value. The index is sorted increasely.
        """
        stripped_pep = self._lib_df['StrippedPeptide'].drop_duplicates()
        mc_ser = stripped_pep.str.count('[KR].*?') - stripped_pep.str.count('.+[KR]$')
        mc_count = mc_ser.value_counts()
        sorted_mc_count = mc_count.sort_index()
        return sorted_mc_count

    def prec_charge_distrib(self) -> pd.Series:
        """
        Count charge in the library
        :return: pandas.Series, which has charge state as index and corresponding peptide number as value. The index is sorted increasely.
        """
        prec_nonredundant_df = self.get_nonredundant_prec_df(self._lib_df)
        prec_charge = prec_nonredundant_df['PrecursorCharge']
        charge_count = prec_charge.value_counts()
        sorted_charge_count = charge_count.sort_index()
        return sorted_charge_count

    def library_ms1_intensity_distrib(self) -> pd.Series:
        """
        MS1 response of each precursor
        :return:
        """
        prec_nonredundant_df = self.get_nonredundant_prec_df(self._lib_df)
        prec_ms1_response = prec_nonredundant_df['ReferenceRunMS1Response']
        return prec_ms1_response

    def library_oxidation_distrib(self):
        """
        :return: Two pd.Series, which are oxidation count and methionine count, respectively
        """
        stripped_pep = self._lib_df['ModifiedPeptide'].drop_duplicates()
        oxi_ser = stripped_pep.str.count(r'M\[Oxi.*?')
        oxi_count = oxi_ser.value_counts()
        sorted_oxi_count = oxi_count.sort_index()
        return sorted_oxi_count

    def library_length_charge_distrib(self) -> pd.Series:
        pass

    def filter_length(self, min_len=None, max_len=None, col='StrippedPeptide'):
        self._lib_df = self._lib_df[self._lib_df[col].str.len().apply(lambda x: min_len <= x <= max_len)]

    def filter_prec_charge(self, charge=(1, 2, 3, 4, 5), prec_charge_col='PrecursorCharge'):
        if isinstance(charge, int):
            charge = (charge,)
        self._lib_df = self._lib_df[self._lib_df[prec_charge_col].astype(int).isin(charge)]

    def filter_peak_losstype(self, min_noloss_peak: int = 6, loss_type='Noloss',
                             prec_col='Precursor', loss_col='FragmentLossType'):
        self._lib_df = self._lib_df.loc[self._lib_df.groupby(prec_col)[loss_col].filter(
            lambda x: x.value_counts().get(loss_type, 0) >= min_noloss_peak).index]

    def filter_library(self, min_len=None, max_len=None, min_noloss_peaks: int = 6, charge=(1, 2, 3, 4, 5)):
        """
        For psm filter, to get get_one_prefix_result library with better quantity for training.
        This will filter the library with three conditions:
            1. stripped peptide length: from min length to 95% max length of whole library length distribution,
            which means the max length range is set as the length when cumulation of sorted pep length num achieves 95% of total pep num.
            The min length is usually set to 7 since the default set in spectronaut is 7.
            The max length can be calculated automatically with ##########################
            2. fragment loss type: 'noloss' type should equals or more than 6 as default
            3. precursor charge: this will be 1-5 in DIA
            Additional condition: 4. miss cleavage: 0-1, this is not so useful

        Either min_len or max_len is not None, the length filter will be done.
        :param min_len: The min length of stripped peptide
        :param max_len: The max length of stripped peptide. An int number will lock the max length to it,
        whill a float number will make the max length an expected number to cover certein percentage of total peptides.
        Example: max_len=26 means the max length is 26, and max_len=0.95 means the length covers 95% of the total peptides will be max length
        :param min_noloss_peaks:
        :param charge:
        """
        if isinstance(charge, int):
            charge = (charge,)

        if isinstance(max_len, float):
            len_distrib = self.strippep_length_distrib()
            if min_len:
                len_distrib = len_distrib[len_distrib.index >= min_len]
            accu_num = 0
            targeted_pep_num = sum(len_distrib) * max_len
            for _l, _n in len_distrib.items():
                accu_num += _n
                if accu_num >= targeted_pep_num:
                    max_len = _l
                    break

        def length_filter(x):
            if min_len:
                if len(x) < min_len:
                    return False
            if max_len:
                if len(x) > max_len:
                    return False
            return True

        length_filtered_df = self._lib_df[self._lib_df['StrippedPeptide'].apply(
            length_filter)] if any((min_len, max_len)) else self._lib_df.copy()  # filter_condition 1

        charge_filtered_df = length_filtered_df[
            length_filtered_df['PrecursorCharge'].astype(int).isin(charge)] if charge else length_filtered_df  # filter_condition 3

        if min_noloss_peaks:
            peak_count = charge_filtered_df.groupby('Precursor')['FragmentLossType'].transform(lambda x: len(x[x == 'noloss']))
            noloss_only_df = charge_filtered_df.loc[peak_count[peak_count >= min_noloss_peaks].index]
            self._lib_df = noloss_only_df
        else:
            self._lib_df = charge_filtered_df

    def library_keep_main_cols(self):
        """
        Remove some columns that are not necessary in the library
        """
        self._lib_df = self._lib_df[SNLibraryTitle.LibraryMainCol]

    def library_keep_main_cols_pgout(self):
        """
        Keep the same columns as library_keep_main_cols except 'ProteinGroups'
        """
        self._lib_df = self._lib_df[SNLibraryTitle.LibraryMainColPGOut]

    def get_prec_list(self):
        if self._prec:
            return list(self._prec)
        else:
            if 'Precursor' not in self._lib_df:
                self.add_prec()
            prec_list = self._lib_df['Precursor'].drop_duplicates().tolist()
            self._prec = prec_list
            return prec_list

    def get_modpep_list(self):
        if self._modpep:
            return list(self._modpep)
        else:
            modpep_list = self._lib_df['ModifiedPeptide'].drop_duplicates().tolist()
            self._modpep = modpep_list
            return modpep_list

    def extract_rt_data(self, pep_col='ModifiedPeptide', rt_col='iRT', return_type='dict'):
        """
        This will extract nonredundancy modpep and corresponding rt
        :return: dict like {'modpep': rt, ...}, or list like [(modpep, rt), ...]
        """
        return get_lib_rt_info(self._lib_df, pep_col, rt_col, return_type)

    def extract_fragment_data(self, norm=False):
        """
        :param norm False means no normalization and a number means the max number of intensity
        """
        frag_dict = get_lib_fragment_info(self._lib_df, norm=norm)
        return frag_dict

    def library_keep_target(self, target_protein_list):
        """
        Keep psms that contain target protein
        target protein may be consist of protein groups, which are separated by semicolons
        :return:
        """
        target_list = rapid_kit.process_list_or_file(target_protein_list)
        target_set = set(target_list)
        self._lib_df = self._lib_df[self._lib_df.apply(rapid_kit.protein_groups_match, col='ProteinGroups', args=(target_set,), axis=1)]

    def library_remove_target(self, target_protein_list):
        """
        Remove psms that contain target protein
        target protein may be consist of protein groups, which are separated by semicolons
        196 ms ± 97.3 µs per loop (mean ± std. dev. of 7 runs, 10 loops each)
        :return:
        """
        target_list = rapid_kit.process_list_or_file(target_protein_list)
        target_set = set(target_list)
        self._lib_df = self._lib_df[~self._lib_df.apply(rapid_kit.protein_groups_match, col='ProteinGroups', args=(target_set,), axis=1)]

    def library_remove_target_re(self, target_protein_list):
        """
        Remove psms that contain target protein
        target protein may be consist of protein groups, which are separated by semicolons
        1min 56s ± 199 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
        Use library_remove_target instead
        """
        if os.path.isfile(target_protein_list):
            with open(target_protein_list, 'r') as target_handle:
                target_list = [_.strip('\n') for _ in target_handle.readlines()]
                while '' in target_list:
                    target_list.remove('')
        else:
            target_list = target_protein_list
        match_target = re.compile('|'.join(['(' + _ + ')' for _ in target_list]))
        self._lib_df = self._lib_df[~(self._lib_df['ProteinGroups'].str.contains(match_target))]

    def check_col_format(self, start_with='_', start_with_num=1, end_with='_', end_with_num=1, normal_char='[A-Z]', special_char=('[', ']'), exclude_char=''):
        pass

    def find_col_format(self):
        pass

    @staticmethod
    def get_nonredundant_prec_df(lib_df: pd.DataFrame) -> pd.DataFrame:
        return lib_df.drop_duplicates(('PrecursorCharge', 'ModifiedPeptide'))

    @staticmethod
    def get_nonredundant_precursor_index(lib_df: pd.DataFrame, position=1) -> pd.Index:
        """
        Consider get_one_prefix_result psm as get_one_prefix_result block.
        :param lib_df:
        :param position: 1 means block start index, -1 means block end index
        :return: pd.Index, which is the first line index or the end line index of each psm block
        """
        precursor_block_index = lib_df.loc[(lib_df['PrecursorCharge'].astype(str) +
                                            lib_df['ModifiedPeptide']) !=
                                           (lib_df['PrecursorCharge'].astype(str).shift(position) +
                                            lib_df['ModifiedPeptide'].shift(position))].index
        return precursor_block_index

    @staticmethod
    def get_psm_block_index(lib_df: pd.DataFrame) -> list:
        """
        Consider get_one_prefix_result psm as get_one_prefix_result block. This can return get_one_prefix_result list contains index tuples for library dataframe slice
        If the first precursor occupy the first 9 lines (which means line 0-8), the tuple will be (0, 9), then library dataframe can be sliced directly.
        :return: list of tuples, like [(0, 9), (9, 13), (13, 19), ...]
        """
        precursor_block_start_index = SpectronautLibrary.get_nonredundant_precursor_index(lib_df, position=1)
        precursor_block_end_index = SpectronautLibrary.get_nonredundant_precursor_index(lib_df, position=-1)
        block_index_list = list(zip(precursor_block_start_index, precursor_block_end_index))
        return block_index_list



import numpy as np


def delta_tx(obse, pred, percent):
    num_x = int(np.ceil(len(obse) * percent))
    return 2 * sorted(abs(np.array(obse) - np.array(pred)))[num_x - 1]


def r_square(obse, pred):
    obse = np.array(obse)
    pred = np.array(pred)
    sse = np.sum(np.square(pred - obse))
    sst = np.sum(np.square(pred - np.mean(obse)))
    r_2 = 1 - sse / sst
    return r_2



import os

"""
文件夹中包含导出的谱图，每个文件按precursor命名
"""


def _num_identi(super_sub_script_num):
    if 8320 <= super_sub_script_num <= 8329:
        return 'super'
    else:
        return 'sub'


def _prm_fragment_name(fragment_content):
    fragment_type = fragment_content[0]
    fragment_num = ''
    fragment_charge = '1'
    for _ in fragment_content[1: -1]:
        ord_num = ord(_)
        num_type = _num_identi(ord_num)
        if num_type == 'super':
            fragment_num += str(ord_num - 8320)
        elif num_type == 'sub':
            fragment_charge = str(ord_num - 176)
    fragment_name = f'{fragment_type}{fragment_num}+{fragment_charge}-noloss'
    return fragment_name


def prm_fragment_process(prm_file_content):
    fragment_intensity_dict = dict()
    for each_byte_intensity_line in prm_file_content[3:]:
        each_intensity_line = each_byte_intensity_line.decode('utf-8')
        split_line = each_intensity_line.strip('\n').split('\t')
        if len(split_line) == 4:
            if 'M' in split_line[2]:
                continue
            each_intensity = float(split_line[1])
            each_fragment = split_line[2]
            if ',' not in each_fragment:
                fragment_name = _prm_fragment_name(each_fragment)
                fragment_intensity_dict[fragment_name] = each_intensity
            else:
                multi_fragments = each_fragment.split(', ')
                fragment_name_list = [
                    _prm_fragment_name(_) for _ in multi_fragments]
                single_electro = [_ for _ in fragment_name_list if '+1' in _]
                if single_electro:
                    fragment_name = single_electro[0]
                else:
                    y_fragment_list = [
                        _ for _ in fragment_name_list if 'y' in fragment_name_list]
                    if y_fragment_list:
                        fragment_name = y_fragment_list[0]
                    else:
                        fragment_name = fragment_name_list[0]
                fragment_intensity_dict[fragment_name] = each_intensity
    return fragment_intensity_dict


def read_prm_result(result_folder):
    prm_intensity_dict = dict()
    for each_file in os.listdir(result_folder):
        each_file_path = os.path.join(result_folder, each_file)
        with open(each_file_path, 'rb') as f:
            file_content = f.readlines()
            each_fragment_dict = prm_fragment_process(file_content)
            precursor = os.path.splitext(each_file)[0]
            charge = precursor[-1]
        prec_info_dict = dict()
        prec_info_dict['Fragment'] = each_fragment_dict
        prec_info_dict['Charge'] = charge
        prec_info_dict['Precursor'] = precursor
        prm_intensity_dict[precursor] = prec_info_dict
    return prm_intensity_dict



"""
1. Heatmap with color blocks (with or without edge lines)
2. With color points not blocks
3. Gap with defined rows or cols
"""

import numpy as np
from matplotlib import pyplot as plt

from mskit.plot_kit.colors.precolors_grad import FavorGradColor
from mskit.plot_kit.drawing_area import remove_xy_ticks


def scatter_heatmap(df, use_row=None, use_col=None,
                    row_name_shift=-3.5, col_name_shift=(-2.5, 3.5), col_name_rotation=45, label_fontsize=8,
                    scatter_thres=None, scatter_func=None, scatter_size=25,
                    anno_fontsize=7, anno_x_shift=0, anno_y_shift=-7.5, anno_format='.2f', anno_style='italic',
                    color_bar=None,
                    grid_color='grey', grid_linewidth=0.5, grid_linestyle='--',
                    ax=None, save=None):
    """
    :param df: The dataframe for ploting. If the df contains some rows or cols not desirable to appear in the figure, use use_row or use_col to control it.
    :param use_row: This can either be used to control the prefered rows to plot or change the order of rows.
    :param use_col: Same as use_row.
    :param row_name_shift: The long should the row name (index of df) shifted from the left axis.
    :param col_name_shift: Similar to row_name_shift but this is a tuple to control the x shift and y shift respectively.
    :param col_name_rotation: The angle to rotate the col name.
    :param label_fontsize: Font size of row names and col names (different from the anno_fontsize below).
    :param scatter_thres: The threshold to control the boundary of data points. This can be None, float or a tuple of two float.
    When this is set to None, all values in df will be retained.
    When this is set to a float, the lower threshold is determined and values below this will be dropped before plotting (set to na actually).
    When this is set to a tuple of two floats, the two boundaries are determined and values beyond this range will be dropped (set to na).
    This will also decide how to scale the values to find a correct color from the given color bar.
    :param scatter_func: A function with a point data input, and output None or False if the data should be dropped.
    :param scatter_size: The area size of the scatter plot.

    :return
    """
    if ax is None:
        ax = plt.gca()
    remove_xy_ticks(ax)
    if not color_bar:
        color_bar = FavorGradColor.DardRed

    if use_row:
        df = df.loc[use_row]
    if use_col:
        df = df.loc[: use_col]

    row_num = df.shape[0]
    col_num = df.shape[1]
    color_num = len(color_bar)

    ax.set_xlim(0, col_num)
    ax.set_ylim(0, row_num)

    scaled_df = df.copy()
    if scatter_thres:
        if len(scatter_thres) == 1:
            scaled_df[scaled_df < scatter_thres] = np.nan
            scaled_df -= scatter_thres
            scaled_df /= scaled_df.values.max()
        elif len(scatter_thres) == 2:
            scaled_df[(scaled_df < scatter_thres[0]) | (scaled_df > scatter_thres[1])] = np.nan
            scaled_df -= scatter_thres[0]
            scaled_df /= (scatter_thres[1] - scatter_thres[0])
    else:
        scaled_df -= scaled_df.values.min()
        scaled_df /= scaled_df.values.max()

    for row_index, (row_name, row_data) in enumerate(df[::-1].iterrows()):
        ax.annotate(row_name, (0, row_index + 0.5),
                    textcoords='offset points', xytext=(row_name_shift, 0),
                    va='center', ha='right', fontsize=label_fontsize, rotation=0)
        for col_index, (col_name, data) in enumerate(row_data.iteritems()):
            if scatter_func:
                if not scatter_func(data):
                    continue
            scaled_data = scaled_df.loc[row_name, col_name]
            if np.isnan(scaled_data):
                continue
            try:
                c = np.array(color_bar[int(scaled_data * color_num)]).reshape(1, 3)
            except IndexError:
                c = np.array(color_bar[int(scaled_data * color_num) - 1]).reshape(1, 3)

            x = col_index + 0.5
            y = row_index + 0.5
            ax.scatter(x, y, c=c, s=scatter_size)
            ax.annotate(format(data, anno_format), (x, y),
                        textcoords='offset points', xytext=(anno_x_shift, anno_y_shift), style=anno_style,
                        va='center', ha='center', fontsize=anno_fontsize, rotation=0)

    for col_index, col_name in enumerate(df.columns):
        ax.annotate(col_name, (col_index + 0.5, row_num),
                    textcoords='offset points', xytext=col_name_shift,
                    va='bottom', ha='left', fontsize=label_fontsize, rotation=col_name_rotation)

    for i in ['xy', 'ij']:
        ax.plot(*np.meshgrid(np.arange(0, col_num + 1), np.arange(0, row_num + 1), indexing=i), c=grid_color, linestyle=grid_linestyle, lw=grid_linewidth)

    if save:
        plt.savefig(save + '.PCC.png')

    return df, scaled_df


def multilayer_heatmap():
    """
    Draw multi row in one block with multi df input with customed bar gap and group gap.
    Fig3
    """
    pass



import re
from typing import Union

from mskit import rapid_kit as rk
from mskit.constants.enzyme import Enzyme

__all__ = [
    'TED',
]


class TED(object):
    def __init__(
            self,
            miss_cleavage=(0, 1, 2),
            min_len=7,
            max_len=33,
            enzyme='Trypsin/P',
            toggle_nterm_m=True,
            return_type='seq',
            extend_n=False
    ):
        """
        TODO exclude some unusual aa if param assigned
        Theoretical Enzyme Digestion -> TED
        :param miss_cleavage: this can be int or tuple, while it will be converted into tuple when use
        :param min_len:
        :param max_len:
        :param enzyme: Enzyme name or a regular expression to define the digestion rules. Currently supported enzymes are trypsin and lysC
        :param toggle_nterm_m: If 'M' on sequence N-terminal, remove this M and go continue (2), both remove and keep this M and go continue (1 or True), nothing to do (0 or False)
        :param return_type: 'seq' or 'site_seq'
        :param extend_n: False or None or int. Nothing to do with the default False, and the n AAs before and after the seq will be returned if int assigned
        """
        self._enzyme = Enzyme()
        self._mc = self._parse_mc(miss_cleavage)
        # TODO Add property for length (min len 7 for default and the func will be able to parse both tuple or int)
        self._min_len = min_len
        self._max_len = max_len
        self._enzyme_names, self._enzyme_rules = self._parse_enzyme(enzyme)
        self._toggle_nterm_m = toggle_nterm_m
        self._return_type = self._parse_return_type(return_type)
        self._extend_n = self._parse_extend_n(extend_n)

    @staticmethod
    def _parse_mc(mc):
        if isinstance(mc, (tuple, list)):
            pass
        elif isinstance(mc, int):
            mc = (mc,)
        else:
            try:
                mc = (int(mc),)
            except TypeError:
                raise TypeError('Miss cleavage shoule be int or tuple of int')
        return mc

    def get_mc(self):
        return self._mc

    def set_mc(self, mc):
        self._mc = self._parse_mc(mc)

    mc = property(get_mc, set_mc, doc='''Miss cleavage for enzyme digestion''')

    def __find_enzyme(self, enzyme, enzyme_names, enzyme_rules):
        if isinstance(enzyme, str):
            if enzyme not in self._enzyme.AllEnzymes:
                print(f'The input enzyme has no pre-defined digestion rule, the enzyme now used is: {enzyme}\n.'
                      f'This input will be regarded as a regular expression.')
                enzyme_rules.append(enzyme)
            else:
                enzyme_names.append(enzyme)
                enzyme_rules.append(self._enzyme.Enzymes[enzyme]['RE'])
        else:
            raise ValueError(f'The input enzyme {enzyme} is not a string')
        return enzyme_names, enzyme_rules

    def _parse_enzyme(self, enzyme):
        enzyme_names = []
        enzyme_rules = []
        if isinstance(enzyme, str):
            enzyme = (enzyme,)
        elif isinstance(enzyme, (list, tuple, set)):
            pass
        else:
            raise ValueError(f'The input enzyme can not be parsed: {enzyme}')
        for each_enzyme in enzyme:
            enzyme_names, enzyme_rules = self.__find_enzyme(each_enzyme, enzyme_names, enzyme_rules)
        return enzyme_names, enzyme_rules

    def get_enzyme(self):
        return self._enzyme_names, self._enzyme_rules

    def set_enzyme(self, enzyme):
        self._enzyme_names, self._enzyme_rules = self._parse_enzyme(enzyme)

    enzyme = property(get_enzyme, set_enzyme, doc='''Enzyme used for digestion''')

    @staticmethod
    def _parse_return_type(return_type: str):
        if return_type not in ['seq', 'site_seq']:
            raise TypeError(f'The input of return_type should be \'seq\' or \'site_seq\', now: {return_type}')
        return return_type

    def get_return_type(self):
        return self._return_type

    def set_return_type(self, return_type: str):
        self._return_type = self._parse_return_type(return_type)

    return_type = property(get_return_type, set_return_type, doc='''Return type can be seq or site_seq.
    If seq: A list of seq will be returned. ['ADEFHK', 'PQEDAK' , ...]
    If site_seq: A list of site and seq will be returned. [(0, 'ADEFHK'), (12, 'PQEDAK'), ...]''')

    @staticmethod
    def _parse_extend_n(extend_n: Union[bool, None, int]):
        if extend_n is True:
            extend_n = 7
            print(f'The extend_n is set to \'True\' and a conventional sequence window 7 is assigned')
        elif extend_n is None:
            extend_n = False
        elif extend_n is False:
            pass
        elif isinstance(extend_n, int):
            if extend_n == 0:
                extend_n = False
            elif extend_n < 0:
                raise ValueError
            else:
                pass
        else:
            raise TypeError(f'The input of extend_n should be False or None or int, now: {extend_n} with a type {type(extend_n)}')
        return extend_n

    def get_extend_n(self):
        return self._extend_n

    def set_extend_n(self, extend_n: Union[bool, None, int]):
        self._extend_n = self._parse_extend_n(extend_n)

    extend_n = property(get_extend_n, set_extend_n, doc='''Extended AA number can be False or None or int.
    Nothing to do with the default False or None is assigned.
    If int: the n AAs before and after the seq will be returned.
    If True: the n will be set to 7 as a conventional sequence window''')

    def __call__(self, seq, add_info=None):
        """
        考虑到之后可能有多种酶同时加入的情况，以及需要前后 n 个 AA 的情况，这里改成对每个 seq 分别用所有 digestion rule 拿到可以切的位点，然后再去组装
            这样可以简化正则的匹配规则同时减少处理步骤
        """
        seq = seq.replace('\n', '').replace(' ', '')
        seq_len = len(seq)
        if self._toggle_nterm_m == 2 and (seq[0] == 'M'):
            seq = seq[1:]

        cleavage_pos = rk.sum_list([[_.end() for _ in re.finditer(rule, seq)] for rule in self._enzyme_rules]) + [0, seq_len]
        if (self._toggle_nterm_m is True or self._toggle_nterm_m == 1) and (seq[0] == 'M'):
            cleavage_pos += [1]
        cleavage_pos = sorted(set(cleavage_pos))

        cleavage_pos_num = len(cleavage_pos)
        pos_comb = []
        for idx, pos in enumerate(cleavage_pos):
            for mc in self._mc:
                end_pos_idx = idx + mc + 1
                if end_pos_idx >= cleavage_pos_num:
                    continue
                else:
                    pos_comb.append((pos, cleavage_pos[end_pos_idx]))

        compliant_seq = []
        for start_idx, end_pos in pos_comb:
            one_seq = seq[start_idx: end_pos]
            seq_len = len(one_seq)
            if seq_len < self._min_len or seq_len > self._max_len:
                continue
            else:
                if self._return_type == 'seq':
                    one_data = (one_seq,)
                elif self._return_type == 'site_seq':
                    one_data = (start_idx, one_seq)
                else:
                    raise ValueError
                if self._extend_n:
                    if start_idx < self._extend_n:
                        prev_seq = '_' * (self._extend_n - start_idx) + seq[: start_idx]
                    else:
                        prev_seq = seq[start_idx - self._extend_n: start_idx]
                    back_seq = seq[end_pos: end_pos + self._extend_n]
                    if len(back_seq) < self._extend_n:
                        back_seq = back_seq + '_' * (self._extend_n - len(back_seq))
                    one_data = (*one_data, prev_seq, back_seq)
                if add_info is not None:
                    if not isinstance(add_info, (tuple, list)):
                        add_info = (add_info,)
                    one_data = (*one_data, *add_info)
                if len(one_data) == 1:
                    one_data = one_data[0]
                compliant_seq.append(one_data)
        return compliant_seq



import datetime
import inspect
import logging
import os
import re
import sys
import shutil

__all__ = [
    'NonOverwriteDict',
    'varname',
    'console_file_logger',
    'load_py_file',
    'recursive_copy',
]


class NonOverwriteDict(dict):
    def __setitem__(self, key, value):
        if self.__contains__(key):
            pass
        else:
            dict.__setitem__(self, key, value)


def varname(var):
    """
    https://stackoverflow.com/questions/592746/how-can-you-print-a-variable-name-in-python
    通过调用这个函数时traceback得到code_content，re到需要的var name
    Traceback(filename='<ipython-input-37-5fa84b05d0d4>', lineno=2, function='<module>', code_context=['b = varname(a)\n'], index=0)
    拿到varname(...)里的内容
    """
    for line in inspect.getframeinfo(inspect.currentframe().f_back)[3]:
        m = re.search(r'\bvarname\s*\(\s*([A-Za-z_][A-Za-z0-9_]*)\s*\)', line)
    if m:
        return m.group(1)


def console_file_logger(
    logger_name='root',
    log_path=None,
    log_console_level=logging.DEBUG,
    log_file_level=None,
    log_console_format='[%(asctime)s] [%(name)s] [%(levelname)s]: %(message)s',
    log_file_format=None,
    write_init_logs=False,
    reset_handlers=True
):
    """
    TODO: custom init messages
    Setup a logger with name `logger_name` and store in `log_path` if path is defined
    """
    logger = logging.getLogger(logger_name)

    if len(logger.handlers) > 0:
        if reset_handlers:
            for h in logger.handlers:
                logger.removeHandler(h)
        else:
            return logger

    start_time = datetime.datetime.now().strftime("%Y%m%d_%H_%M_%S")
    stored_msgs = [
        f'Logger "{logger_name}" created at {start_time}',
        f'Logger "{logger_name}" has level (console): "{log_console_level}"',
        f'Logger "{logger_name}" has format (console): "{log_console_format}"',
    ]

    logger.setLevel(log_console_level)

    formatter = logging.Formatter(log_console_format)

    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    stdout_handler.setLevel(log_console_level)
    stdout_handler.setFormatter(formatter)
    logger.addHandler(stdout_handler)

    if log_path is not None:
        file_handler = logging.FileHandler(log_path)

        if log_file_level is None:
            log_file_level = log_console_level
        file_handler.setLevel(log_file_level)

        if log_file_format is None:
            log_file_format = log_console_format
            formatter = logging.Formatter(log_file_format)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        stored_msgs.extend([
            f'Set store path of logger {logger_name} to: {log_path}'
            f'Logger "{logger_name}" has level (file): "{log_file_level}"'
            f'Logger "{logger_name}" has format (file): "{log_file_format}"'
        ])

    if write_init_logs:
        for msg in stored_msgs:
            logger.info(msg)

    return logger


def load_py_file(file_path, remove_insert_path='new'):
    """
    :param file_path:

    :param remove_insert_path:
        'any' to remove inserted path even the path is in path list before performing this function
        'new' (default) to remove inserted path from sys path list if this path was not in list before
        False to do nothing
    """
    # TODO remove inserted path?
    file_dir = os.path.dirname(file_path)
    file_name = os.path.basename(file_path)
    file_name_no_suffix = os.path.splitext(os.path.basename(file_name))[0]
    sys.path.insert(-1, file_dir)
    content = {}
    try:
        exec(f'import py file {file_name}', {}, content)
    except ModuleNotFoundError:
        raise FileNotFoundError(f'Not find input file {file_name} with basename {file_name_no_suffix} in {file_dir}')
    return content['content']


def recursive_copy(original, target, ignored_items=None, verbose=True, exist_ok=True):
    if ignored_items is None:
        ignored_items = []

    os.makedirs(target, exist_ok=exist_ok)
    curr_items = os.listdir(original)
    for item in curr_items:
        if item in ignored_items:
            continue
        original_item_path = os.path.join(original, item)
        target_item_path = os.path.join(target, item)
        if os.path.isdir(original_item_path):
            recursive_copy(original_item_path, target_item_path, ignored_items=ignored_items, verbose=verbose)
        elif os.path.isfile(original_item_path):
            if verbose:
                print(f'copying {item} from {original_item_path} to {target_item_path}')
            shutil.copy(original_item_path, target_item_path)
        else:
            raise
    return 0



import numpy as np
from scipy.stats import pearsonr

from mskit.metric import similarity

"""
Use all label / Use shared
From array / From dict
array 或 dict 的上一级，prec 水平
返回值包含所有 prec union 的数量，shared 数量，即两个 n
删除完全没有匹配或给一个默认值 0？
"""

"""
1. One side: 一边为 benchmark，另一边缺失补零 or other number
2. Two side: 两边并集，缺失补零
3. Shared: 两边共有
"""


def inten_pair_from_dict(inten1: dict, inten2: dict, method='OneSide', fill_value=0):
    """
    # Get the two intensity lists from input inten dicts and the values are remained through the keys in main_dict
    # The missing values in test_dict will be filled with zero as default
    Input dicts with key-value pairs as [frag_name, intensity]
    """
    inten1_list, inten2_list = [], []
    if method == 'Shared':
        frags = inten1.keys() & inten2.keys()
    elif method == 'OneSide':
        frags = inten1.keys()
    elif method == 'TwoSide':
        frags = inten1.keys() | inten2.keys()
    else:
        raise ValueError(f'The get_frag_list function requires a method param with through_main or shared. {method} is passed now.')
    for frag in frags:
        inten1_list.append(inten1.get(frag, fill_value))
        inten2_list.append(inten2.get(frag, fill_value))
    return inten1_list, inten2_list


def inten_pair_from_array(main_array: np.ndarray, test_array: np.ndarray, filter_func=None):
    """
    Get two intensity pair lists from input two inten array and the remained values are followed
    with the idx of values within the range of defined min_num and max_num in main_array
    Example of filter_func:
        filter_func=lambda x: (x > 0) & (x <= 1)
    """
    if main_array.shape != test_array.shape:
        return None
    main_array = main_array.reshape(-1)
    test_array = test_array.reshape(-1)
    if filter_func:
        used_idx = np.where(filter_func(main_array))
    else:
        used_idx = np.where(main_array)
    return main_array[used_idx], test_array[used_idx]


def get_frag_list(fragment_data1, fragment_data2, method='OneSide'):
    matched_pair_dict = dict()
    shared_prec = fragment_data1.keys() & fragment_data2.keys()
    for prec in shared_prec:
        inten_data1 = fragment_data1[prec]
        inten_data2 = fragment_data2[prec]
        data_list1, data_list2 = inten_pair_from_dict(inten_data1, inten_data2, method=method, fill_value=0)
        matched_pair_dict[prec] = (data_list1, data_list2)
    return matched_pair_dict


def calc_pcc(data1, data2, keep_pvalue=False):
    pcc = pearsonr(data1, data2)
    if keep_pvalue:
        return pcc
    else:
        return pcc[0]


def calc_sa(data1, data2):
    data1 = np.array(data1)
    data2 = np.array(data2)
    norm_data1 = data1 / np.sqrt(sum(np.square(data1)))
    norm_data2 = data2 / np.sqrt(sum(np.square(data2)))
    sa = 1 - 2 * (np.arccos(sum(norm_data1 * norm_data2))) / np.pi
    return sa


def calc_pcc_with_dictdata(data1, data2, method='OneSide', fill_value=0, na_value=None):
    pcc = similarity.pcc(*inten_pair_from_dict(data1, data2, method=method, fill_value=fill_value))
    if na_value is not None and np.isnan(pcc):
        pcc = na_value
    return pcc


def calc_sa_with_dictdata(data1, data2, method='OneSide'):
    return similarity.sa(*inten_pair_from_dict(data1, data2, method=method, fill_value=0))


def frag_pcc(main_frag_data, test_frag_data, min_pairs=3, keep_type='through_main', keep_pvalue=False):
    pcc_dict = dict()
    matched_pair_dict = get_frag_list(main_frag_data, test_frag_data, method=keep_type)
    for prec, (main_list, test_list) in matched_pair_dict.items():
        if len(main_list) < min_pairs or len(test_list) < min_pairs:
            continue
        pcc = calc_pcc(main_list, test_list, keep_pvalue=keep_pvalue)
        if np.isnan(np.min(pcc)):
            continue
        pcc_dict[prec] = pcc
    return pcc_dict


def frag_sa(main_frag_data, test_frag_data, min_pairs=3, keep_type='through_main'):
    sa_dict = dict()
    matched_pair_dict = get_frag_list(main_frag_data, test_frag_data, method=keep_type)
    for prec, (main_list, test_list) in matched_pair_dict:
        if len(main_list) < min_pairs or len(test_list) < min_pairs:
            continue
        sa = calc_sa(main_list, test_list)
        sa_dict[prec] = sa
    return sa_dict


def write_pcc_result(pred_file_path, pcc_dict):
    with open(r'{}.fragment.info'.format(pred_file_path), 'w') as f:
        for _, __ in pcc_dict.items():
            f.write('{}\t{}\n'.format(_, __))


def write_sa_result(pred_file_path, pcc_dict):
    with open(r'{}.sa.info'.format(pred_file_path), 'w') as f:
        for _, __ in pcc_dict.items():
            f.write('{}\t{}\n'.format(_, __))



"""
    Comet style (modifications with symbols annotated):
        S@/T@/Y@ for Phospho (STY)
        M* for Oxidation (M)
        n# for Acetyl (N-term)
        Example: n#DFM*SPKFS@LT@DVEY@PAWCQDDEVPITM*QEIR
"""


def comet_to_intseq(x):
    x = x.replace('_', '')

    x = x.replace('M*', '1')
    x = x.replace('S@', '2')
    x = x.replace('T@', '3')
    x = x.replace('Y@', '4')

    if 'n#' in x:
        x = x.replace('n#', '')
        x = '*' + x
    else:
        x = '@' + x

    return x



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.backends.backend_pdf import PdfPages


def save_pdf(path):
    with PdfPages(path) as pdf:
        pdf.savefig()


def to_array(figure=None):
    if figure is None:
        figure = plt.gcf()
    canvas = FigureCanvasAgg(figure)
    canvas.draw()
    image_array = np.array(canvas.renderer.buffer_rgba())
    return image_array




_MAJOR_VERSION = 0
_MINOR_VERSION = 4
_MICRO_VERSION = 1
# _RELEASE_DATE = '.'
# _RELEASE_TYPE = 'alpha'
__version__ = '.'.join(
    map(str,
        [_MAJOR_VERSION, _MINOR_VERSION, _MICRO_VERSION])
)

__package_name__ = 'mskit'
__author__ = 'Lou, R.'
__email__ = 'lourh@shanghaitech.edu.cn'
__repository__ = 'https://github.com/gureann/MSKit'
__license__ = 'MIT'

__short_description__ = 'A tool-kit for MS data processing'



import numpy as np
import scipy.stats
import sklearn


def iqr(value: np.ndarray):
    return np.percentile(value, 75) - np.percentile(value, 25)


def cv(value_array: np.ndarray, min_quant_value_num=3, std_ddof=1, make_percentage=True, keep_na=False, decimal_place=None, return_iqr=False):
    """
    :param value_array: A two-dimensional array with rows as sample and cols as replicates. CV will be performed to each row (dim 0)
    :param min_quant_value_num: Minimum number of non-NA values
    :param std_ddof: ddof for std
    :param make_percentage: If true, CVs will be multi with 100, else nothing will do. Default True
    :param keep_na: Whether to return NAs for those CV-unavailable samples. If True, the returned CVs will have the same size as input samples. Default False
    :param decimal_place:
    :param return_iqr: Whether to return IQR of calulated CVs. If True, a tuple like (cvs, iqr) will be returned, otherwise cvs only. Default False
    """
    if len(value_array.shape) != 2:
        raise ValueError(f'Calc CV would expects a two-dim array with sample as rows and replicates as cols. Current input array has {len(value_array.shape)} dim')
    sample_num, rep_num = value_array.shape
    if min_quant_value_num > rep_num:
        min_quant_value_num = rep_num
    cv_avail_value_idx = np.where((rep_num - np.isnan(value_array).sum(axis=1)) >= min_quant_value_num)[0]
    cv_avail_values = value_array[cv_avail_value_idx]
    cvs = np.nanstd(cv_avail_values, axis=1, ddof=std_ddof) / np.nanmean(cv_avail_values, axis=1)
    if make_percentage:
        cvs = cvs * 100
    if keep_na:
        temp = np.zeros(sample_num)
        temp.fill(np.nan)
        temp[cv_avail_value_idx] = cvs
        cvs = temp.copy()
    return cvs


def count_missing_values(value_array: np.ndarray, keep_all_na_row=True):
    mvs = np.sum(np.isnan(value_array), axis=1)
    return mvs


def fwhm(values: np.ndarray, est_x_num=1e3):
    """
    :return: A tuple as (FWHM value, APEX point, Max estimated Y, KDE func)
    """
    sorted_values = np.sort(values)
    kde = scipy.stats.kde.gaussian_kde(sorted_values)
    kde_x = np.linspace(sorted_values[0], sorted_values[-1], int(est_x_num))
    kde_y = kde(kde_x)
    max_est_y = np.max(kde_y)
    apex_x_idx = np.argmax(kde_y)
    fwhm_min_idx = np.where(kde_y[:apex_x_idx] < kde_y[apex_x_idx] / 2)[0][-1]
    fwhm_max_idx = np.where(kde_y[apex_x_idx:] < kde_y[apex_x_idx] / 2)[0][0] + apex_x_idx
    apex_point = kde_x[apex_x_idx]
    fwhm_value = kde_x[fwhm_max_idx] - kde_x[fwhm_min_idx]
    return fwhm_value, apex_point, max_est_y, kde


# def pca():
#     cond_quant_df = quant_df[conditions].dropna(how='any').copy()
#     pca_input_values = cond_quant_df.values.T
#     pca = sklearn.decomposition.PCA(n_components=2).fit(pca_input_values)
#     component_var = pca.explained_variance_ratio_
#     transformed_values = pca.transform(pca_input_values)



import numpy as np

from .trans_color_code import rgb_to_hsl, hsl_to_rgb


def gradient_hsl(start, end, number, input_color_type='rgb', output_color_type='rgb', value_scale=None):
    if number < 2:
        raise
    if input_color_type.lower() != 'rgb':
        raise

    colors = np.array([
        np.linspace(*each_channel, number) for each_channel in zip(
            rgb_to_hsl(start), rgb_to_hsl(end))]
    ).transpose()

    if output_color_type.lower() == 'rgb':
        colors = [hsl_to_rgb(_) for _ in colors]
    if value_scale:
        colors = [[__ / value_scale for __ in _] for _ in colors]
    return colors


def gradient_rgb(start, end, number, input_color_type='rgb', output_color_type='rgb', value_scale=None):
    if number < 2:
        raise
    if input_color_type.lower() != 'rgb':
        raise

    colors = np.array([
        np.linspace(*each_channel, number) for each_channel in zip(start, end)]
    ).transpose()
    colors = colors.astype(int)
    if value_scale:
        colors = [[__ / value_scale for __ in _] for _ in colors]
    return colors




from ._spectronaut import SpectronautLibrary

from ._spectronaut import sn_constant
from ._spectronaut import sn_utils
from ._spectronaut.sn_utils import *
from ._spectronaut import mod_process
from ._spectronaut import sn_result



import matplotlib.pyplot as plt


def isometric_axes(
        left_init=0.05, bottom_init=0.05, right_end=0.9, top_end=0.9,
        ax_col_gap=0.1, ax_row_gap=0.1,
        row_num=2, col_num=5,
        total_num=None,
        container=None,
        *figure_args
):  # TODO support new ax args
    if container is None:
        container = plt.figure(*figure_args)
    if isinstance(container, plt.Figure):
        new_ax_method = container.add_axes
    elif isinstance(container, plt.Axes):
        new_ax_method = container.inset_axes
    else:
        raise ValueError(f'container should be `Figure` or `Axes` or `None`, now {type(container)}')

    if isinstance(ax_col_gap, (float, int)):
        ax_col_gap = [ax_col_gap] * (col_num - 1)
    if isinstance(ax_row_gap, (float, int)):
        ax_row_gap = [ax_row_gap] * (row_num - 1)
    else:
        ax_row_gap = ax_row_gap[::-1]
    if not total_num:
        total_num = row_num * col_num

    fig_length = right_end - left_init
    fig_height = top_end - bottom_init

    row_height = (fig_height - sum(ax_row_gap)) / row_num
    col_length = (fig_length - sum(ax_col_gap)) / col_num

    ax_x = [left_init + sum(ax_col_gap[:i]) + col_length * i for i in range(col_num)]
    ax_y = [top_end - sum(ax_row_gap[:i]) - row_height * (i + 1) for i in range(row_num)]

    ax_num = 0
    axes_list = []
    for row_index, y in enumerate(ax_y):
        for col_index, x in enumerate(ax_x):
            if ax_num == total_num:
                break
            ax = new_ax_method([x, y, col_length, row_height])
            axes_list.append(ax)
            ax_num += 1

    return container, axes_list



from matplotlib import pyplot as plt
from mskit.stats import delta_tx, r_square
from scipy.stats import pearsonr

from mskit.plot_kit.colors.preferred_colors import S

# TODO Reg class
"""
将绘制函数放在 class 中还是独立的函数中？

class XXX:
    def __init__(xxxxx)
        self.figure = None
        self.ax = None
        self.data = None
    def draw(xxxxx):
        
        
def rt_reg(Other_Params=None, ..., XXX=None):
    if XXX=None:
        x = XXX(....)
        x.draw()
        return x
    else:
        x = XXX
        update x param
        x.draw()
        return x
"""


def define_reg_ax(ax, rt_1, rt_2, x_label, y_label, rt_unit, expand=0.2):
    rt_range = (min(min(rt_1), min(rt_2)), max(max(rt_1), max(rt_2)))
    rt_width = rt_range[1] - rt_range[0]
    min_axis = rt_range[0] - rt_width * expand
    max_axis = rt_range[1] + rt_width * expand
    ax.set_xlim(min_axis, max_axis)
    ax.set_ylim(min_axis, max_axis)
    ax.set_aspect('equal')
    if rt_unit:
        unit_suffix = f' ({rt_unit})'
    else:
        unit_suffix = ''
    ax.set_xlabel(f'{x_label}{unit_suffix}')
    ax.set_ylabel(f'{y_label}{unit_suffix}')
    return min_axis, max_axis


def ex_line_dict(obse_rt, pred_rt, ex_line):
    ex_dict = dict()
    ex_line = sorted(list(ex_line), reverse=True)
    for _each_ex in ex_line:
        _each_ex = _each_ex * 100 if _each_ex <= 1 else _each_ex
        ex_dict[str(_each_ex)] = delta_tx(obse_rt, pred_rt, _each_ex / 100)
    return ex_dict


def rt_reg(obse_rt, pred_rt,
           ex_line=(95, 85, 75), ex_line_num=1, ex_linewidth=1., diagonal=True, diagonal_linewidth=1.,
           scatter_size=1.5, scatter_color='#5250AD',
           rt_unit='units',  # rt_units -> label_suffix
           anno_fontsize=10, anno_gap=12.5, anno_group_row_gap: int = 1,  # anno text, unit of anno number
           # PCC:True/False, R2:True/False, n:True/False, combine these params to a dict and the order is defined as the key order in dict
           # How to add other params    active=False -> Return the class and need to run .active_anno() method
           x_label='Observed RT', y_label='Predicted RT', title='RT correlation',
           scat_label=None,
           ax=None, save=None):
    if ax is None:
        ax = plt.gca()
    min_axis, max_axis = define_reg_ax(ax, obse_rt, pred_rt, x_label, y_label, rt_unit)

    if ex_line:
        ex_dict = ex_line_dict(obse_rt, pred_rt, ex_line)
    else:
        ex_dict = None

    ax.set_title(title)

    scat = ax.scatter(obse_rt, pred_rt, c=scatter_color, s=scatter_size, label=scat_label)

    if diagonal:
        ax.plot([min_axis, max_axis], [min_axis, max_axis], color='k', linewidth=diagonal_linewidth)

    annotate_text_ex = []
    if ex_dict:
        for _i, (_ex_num, _ex_value) in enumerate(ex_dict.items()):
            try:
                _color = S[_i]
            except IndexError:
                _color = 'k'
            if _i < ex_line_num:
                ax.plot([min_axis, max_axis], [min_axis - _ex_value / 2, max_axis - _ex_value / 2], color=_color, linewidth=ex_linewidth)
                ax.plot([min_axis, max_axis], [min_axis + _ex_value / 2, max_axis + _ex_value / 2], color=_color, linewidth=ex_linewidth)
            annotate_text_ex.append(rf'$\Delta$t$_{{{_ex_num}\%}}$ = {_ex_value:.3f} ({rt_unit})')

    p = pearsonr(obse_rt, pred_rt)
    r_2 = r_square(obse_rt, pred_rt)

    annotate_text_info = [f'PCC: {p[0]:.4f}', f'R$^2$: {r_2:.4f}', f'n={format(len(obse_rt), ",")}']

    annotation = annotate_text_ex + [''] * anno_group_row_gap + annotate_text_info
    for _i, _anno in enumerate(annotation):
        _ = ax.annotate(_anno, xy=(min_axis, max_axis), fontsize=anno_fontsize,
                        va='top', ha='left',
                        textcoords='offset points', xytext=(5, -anno_gap * (_i + 0.5)))

    if save:
        plt.savefig(save + '.RT.png')

    return scat, annotate_text_ex, annotate_text_info



from . import intensity
from . import mass_calc
from .intensity import normalize_intensity, keep_top_n_inten
from .mass_calc import calc_prec_mz, calc_fragment_mz, calc_fragment_mz_old, get_fragment_mz_dict



import os
from sys import argv

import numpy as np
import pandas as pd
import scipy.stats


def read_need_list(list_content):
    if os.path.isfile(list_content):
        with open(list_content, 'r') as f:
            need_list = f.read().split('\n')
    elif isinstance(list_content, str):
        need_list = list_content.split('\n')
    else:
        need_list = list_content
    need_list = [_.strip('\n') for _ in need_list]
    while '' in need_list:
        need_list.remove('')
    need_list = [_.replace('C', 'C[+57.021464]') for _ in need_list]
    return need_list


def process_transition_file(
    transition_file,
    need_list: list = None,
    kept_fragment_name: tuple = ('b1', 'b2', 'y1', 'y2'),
) -> pd.DataFrame:
    transition_df = pd.read_csv(transition_file, header=-1)
    if need_list is not None:
        transition_df = transition_df[transition_df[3].isin(need_list)]

    transition_df = transition_df.drop(transition_df.loc[transition_df[5] == transition_df[5].shift(1)].index)
    transition_df = transition_df[~(transition_df[5].isin(kept_fragment_name))]
    return transition_df


if __name__ == '__main__':
    trans_file = argv[1]
    nl = read_need_list(argv[2])
    trans_df = process_transition_file(trans_file, nl)
    out_path = os.path.splitext(trans_file)[0] + '-Processed.csv'
    trans_df.to_csv(out_path, index=False, header=False)


def transition_list_filter(raw_file, pep_list):
    pass


def read_skyline_result(sky_file, partial_run_names=None):
    result_df = pd.read_csv(sky_file)
    stripped_pep_list = result_df['Peptide Sequence'].drop_duplicates().dropna().tolist()
    sample_list = result_df['Replicate Name'].drop_duplicates().dropna().tolist()
    df_columns = result_df.columns

    if 'Area' not in df_columns:
        use_total_fragment_area = True
    else:
        use_total_fragment_area = False

    if partial_run_names:
        sample_list = (
                [_ for _ in sample_list if partial_run_names[0].lower() in _.lower()]
                + [_ for _ in sample_list if partial_run_names[1].lower() in _.lower()]
        )
    sample_num = len(sample_list)
    pep_num = len(stripped_pep_list)
    quant_df = pd.DataFrame(np.zeros((pep_num, sample_num)), index=stripped_pep_list, columns=sample_list)

    for each_stripep in stripped_pep_list:
        each_pep_df = result_df[result_df['Peptide Sequence'] == each_stripep]
        each_prec_df = each_pep_df[each_pep_df['Fragment Ion'] == 'precursor']
        each_fragment_df = each_pep_df[~(each_pep_df['Fragment Ion'] == 'precursor')]

        for each_rep in sample_list:
            if use_total_fragment_area:
                each_fragment_area_sum = \
                each_prec_df[each_prec_df['Replicate Name'] == each_rep]['Total Area Fragment'].iloc[0]
            else:
                try:
                    coeluting_true_df = each_fragment_df[each_fragment_df['Coeluting'] == True]
                except KeyError:
                    coeluting_true_df = each_fragment_df
                each_rep_df = coeluting_true_df[coeluting_true_df['Replicate Name'] == each_rep]
                try:
                    each_fragment_area_sum = each_rep_df['Area'].sum()
                except:
                    raise
            quant_df.loc[each_stripep, each_rep] = each_fragment_area_sum

    for _ in range(sample_num):
        quant_df.iloc[:, _] /= align_list[_]
    return quant_df


def drop_rep(df, method='StandardScore'):
    samp_list = df.columns
    sample_num = len(samp_list)
    rep_num = int(sample_num / 2)

    df1 = df.loc[:, samp_list[0: rep_num]]
    df2 = df.loc[:, samp_list[rep_num: rep_num * 2]]

    df1 = df1.drop(calc_df_ss_sum(df1).idxmax(), axis=1)
    df2 = df2.drop(calc_df_ss_sum(df2).idxmax(), axis=1)
    return pd.merge(df1, df2, left_index=True, right_index=True)


def calc_df_ss_sum(df):
    """
    增加参数：
        剩余的rep数量
        每个肽段的权重
    """

    def calc_standard_score(x):
        m = np.mean(x)
        s = np.std(x, ddof=1)
        _ = x - m
        _ = _ / s
        return _

    standard_score_df = df.apply(calc_standard_score, axis=1).applymap(abs)
    ss_sum = standard_score_df.apply(lambda x: sum([_ for _ in x if _ != x.max()]), axis=0)
    return ss_sum


def calc_all_ratio(df1, df2, method='mean'):
    if df1.shape != df2.shape:
        raise
    nrow, ncol = df1.shape
    ratio_matrix = np.zeros((nrow, ncol ** 2))
    for row_num in range(nrow):
        col_num = 0
        for col_num1 in range(ncol):
            for col_num2 in range(ncol):
                ratio_matrix[row_num, col_num] = df2.iloc[row_num, col_num2] / df1.iloc[row_num, col_num1]
                col_num += 1
    return ratio_matrix.mean(axis=1), ratio_matrix.std(axis=1, ddof=1)


def stats_within_group(quant_df, sample_identi):
    samp_list = quant_df.columns
    sample_num = len(samp_list)
    rep_num = int(sample_num / 2)

    # MeanArea
    quant_df[f'MeanArea-{sample_identi[0]}'] = quant_df.loc[:, samp_list[0: rep_num]].apply(np.mean, axis=1)
    quant_df[f'MeanArea-{sample_identi[1]}'] = quant_df.loc[:, samp_list[rep_num: rep_num * 2]].apply(np.mean, axis=1)
    # Ratio-respective
    for rep_series in range(rep_num):
        each_bi = quant_df.iloc[:, rep_series + rep_num] / quant_df.iloc[:, rep_series]
        quant_df[f'Ratio-{rep_series + 1}'] = each_bi
    # Ratio-Mean
    quant_df[f'Ratio-Mean'] = quant_df.loc[:, [_ for _ in quant_df.columns if 'Ratio' in _]].apply(np.mean, axis=1)
    # STD-Ratio
    quant_df[f'STD-Ratio'] = quant_df.loc[:, [_ for _ in quant_df.columns if 'Ratio' in _]].apply(np.std, ddof=1,
                                                                                                  axis=1)
    # CV-Ratio
    quant_df[f'CV-Ratio'] = quant_df[f'STD-Ratio'] / quant_df[f'Ratio-Mean']
    # p-value
    pvalue_list = []
    for each_pep in quant_df.index:
        if scipy.stats.levene(
                quant_df.loc[each_pep, samp_list[0: rep_num]],
                quant_df.loc[each_pep, samp_list[rep_num: rep_num * 2]]
        )[1] < 0.05:
            equal_var = False
        else:
            equal_var = True
        t_result = scipy.stats.ttest_ind(
            quant_df.loc[each_pep, samp_list[0: rep_num]],
            quant_df.loc[each_pep, samp_list[rep_num: rep_num * 2]],
            equal_var=equal_var)
        pvalue_list.append(t_result[1])
    quant_df[f'P-value'] = pvalue_list
    # MeanRatio
    mean_multi_ratio, std_multi_ratio = calc_all_ratio(quant_df.loc[:, samp_list[0: rep_num]],
                                                       quant_df.loc[:, samp_list[rep_num: rep_num * 2]])
    quant_df['MeanRatio'] = mean_multi_ratio
    quant_df['STD-MeanRatio'] = std_multi_ratio
    # STD
    quant_df[f'STD-{sample_identi[0]}'] = quant_df.loc[:, samp_list[0: rep_num]].apply(np.std, ddof=1, axis=1)
    quant_df[f'STD-{sample_identi[1]}'] = quant_df.loc[:, samp_list[rep_num: rep_num * 2]].apply(np.std, ddof=1, axis=1)
    # CV
    quant_df[f'CV-{sample_identi[0]}'] = quant_df[f'STD-{sample_identi[0]}'] / quant_df[
        f'MeanArea-{sample_identi[0]}'] * 100
    quant_df[f'CV-{sample_identi[1]}'] = quant_df[f'STD-{sample_identi[1]}'] / quant_df[
        f'MeanArea-{sample_identi[1]}'] * 100

    return quant_df


def write_result(input_file, quant_df, quant_df2=None):
    out_path = os.path.splitext(input_file)[0] + '-ProcessedAligned.xlsx'

    with pd.ExcelWriter(out_path) as f:
        quant_df.style.applymap(lambda _: 'font-family: Aria').to_excel(f, sheet_name='RawData')
        quant_df2.style.applymap(lambda _: 'font-family: Aria').to_excel(f, sheet_name='RepDrop')


if __name__ == '__main__':
    result_file = r'.\Result.csv'
    sample_identifier = 'Control,Experiment'
    align_list = [1.1, 1.1, 1.1, 1.1] + [1.1, 1.1, 1.1, 1.1]
    sample_identifier = sample_identifier.split(',')
    raw_result_df = read_skyline_result(result_file, sample_identifier)
    rep_drop_df = drop_rep(raw_result_df)
    processed_result_df = stats_within_group(raw_result_df, sample_identifier)
    processed_result_drop_df = stats_within_group(rep_drop_df, sample_identifier)
    write_result(result_file, processed_result_df, processed_result_drop_df)



import re

from .data_struc_kit import str_mod_to_list


def split_str_with_symbol_included_substring(
        s: str,
        split_symbol: str = '\t',
        special_symbol: str = '"',
        keep_spcial_symbol: bool = False,
        strip_slash_n: bool = False
):
    if strip_slash_n:
        s = s.strip('\n')
    split_str = []
    symbol_start = False
    substr = ''
    for c in s:
        if c == special_symbol:
            symbol_start = [True, False][symbol_start]
            if keep_spcial_symbol:
                substr += c
        elif c == split_symbol:
            if symbol_start:
                substr += c
            else:
                split_str.append(substr)
                substr = ''
        else:
            substr += c
    split_str.append(substr)
    return split_str


def get_title2idx_dict(title_content: str) -> dict:
    title_dict = dict([(__, _) for _, __ in enumerate(
        title_content.strip('\n').split('\t'))])
    return title_dict


def combine_delimited_text(
        *sn: str,
        delimiter: str = ';',
        keep_order: bool = False
) -> str:
    """
    This will combine two strings with semicolons and drop duplication
    Example: s1='Q1;Q2', s2='Q2;Q3' -> 'Q1;Q2;Q3'
    Note that the order may change if keep_order=False
    """
    s_list = map(lambda _: _.strip(delimiter).split(delimiter), sn)
    flatten_s = list(filter(lambda x: True if x else False, sum(s_list, [])))
    unique_s = list(set(flatten_s))
    if keep_order:
        unique_s = sorted(unique_s, key=flatten_s.index)
    return ';'.join(unique_s)


def find_substring(
        s: str,
        start_char: str = '[',
        end_char: str = ']',
        keep_start_end_char: bool = True,
        substring_trans_dict: dict = None
):
    """
    This will find the substrings start with param "start" and end with param "end" in input "string"
    This will return
        a string with no substrings
        a list of positions of substrings (the 1-indexed)
        a list of substrings
    Example:
        substring_finder('AM(ox)M(Oxidation (M))C(Carbamidomethyl (C))DEEHC(Carb)K', '(', ')')
            ('AMMCDEEHCK',
            [2, 3, 4, 9],
            ['(ox)', '(Oxidation (M))', '(Carbamidomethyl (C))', '(Carb)'])
        substring_finder('AM[ox]M[Oxidation (M)]C[Carbamidomethyl (C)]DEEHC[Carb]K', '[', ']')
            ('AMMCDEEHCK',
            [2, 3, 4, 9],
            ['[ox]', '[Oxidation (M)]', '[Carbamidomethyl (C)]', '[Carb]'])
    TODO 再返回一个对应位点氨基酸的 list  注意 N 和 C 端
    TODO substring_trans_dict
    """
    start_num = 0
    end_num = 0
    substrings = []
    pos = []
    substring_start = False
    sub_total_len = 0
    main_str = ''
    sub_str = ''
    for i, char in enumerate(s):
        if char == start_char:
            sub_str += char
            substring_start = True
            start_num += 1
        elif char == end_char:
            sub_str += char
            end_num += 1
            if start_num == end_num:
                substrings.append(sub_str)
                sub_total_len += len(sub_str)
                pos.append(i - sub_total_len + 1)
                start_num = 0
                end_num = 0
                sub_str = ''
                substring_start = False
        else:
            if substring_start:
                sub_str += char
            else:
                main_str += char
    return main_str, pos, substrings


substring_finder = find_substring


def fillin_annotation(
        s,
        anno_pos,
        anno_text,
        existed_anno_start_char: str = None,
        existed_anno_end_char: str = None
) -> str:
    if isinstance(anno_pos, (str, int)):
        anno_pos = [int(anno_pos)]
    if isinstance(anno_text, str):
        anno_text = [anno_text]

    if existed_anno_start_char is not None and existed_anno_end_char is not None:
        s, exist_anno_pos, exist_anno_text = find_substring(s, start_char=existed_anno_start_char, end_char=existed_anno_end_char)
    elif existed_anno_start_char is None and existed_anno_end_char is None:
        exist_anno_pos, exist_anno_text = [], []
    else:
        raise ValueError(
            'Both `existed_anno_start_char` and `existed_anno_end_char` need to be passed or set to None as default. '
            f'Now {existed_anno_start_char} and {existed_anno_end_char}'
        )

    exist_anno_pos += anno_pos
    exist_anno_text += anno_text

    t = sorted([(i, v) for i, v in enumerate(exist_anno_pos)], key=lambda x: x[1])
    anno_pos = [_[1] for _ in t]
    anno_text = [anno_text[i] for i in [_[0] for _ in t]]
    anno_s = ''
    _ = 0
    for pos_idx, pos in enumerate(anno_pos):
        anno_s += s[_: pos]
        anno_s += anno_text[pos_idx]
        _ = pos
    anno_s += s[pos:]
    return anno_s


def extract_bracket(str_with_bracket, ):  # Need a parameter to choose to use () or [] or others (by manurally define?) and a parameter to skip how many additional brackets
    bracket_start = [left_bracket.start() for left_bracket in re.finditer('\(', str_with_bracket)]  # Add [::2] if there is one additional bracket in the expected one
    bracket_end = [right_bracket.start() for right_bracket in re.finditer('\)', str_with_bracket)]  # Add [1::2] if add the additional operation at the last step
    return bracket_start, bracket_end


def split_fragment_name(fragment_name):
    frag_type, frag_num, frag_charge, frag_loss = re.findall(
        '([abcxyz])(\\d+)\\+(\\d+)-?(.*)', fragment_name)[0]
    return frag_type, int(frag_num), int(frag_charge), frag_loss


def split_prec(prec: str, keep_underline=False):
    modpep, charge = prec.split('.')
    if not keep_underline:
        modpep = modpep.replace('_', '')
    return modpep, int(charge)


def assemble_prec(modpep, charge):
    if not modpep.startswith('_'):
        modpep = f'_{modpep}_'
    return f'{modpep}.{charge}'


def split_mod(modpep, mod_ident='bracket'):
    if mod_ident == 'bracket':
        mod_ident = ('[', ']')
    elif mod_ident == 'parenthesis':
        mod_ident = ('(', ')')
    else:
        pass
    re_find_pattern = '(\\{}.+?\\{})'.format(*mod_ident)
    re_sub_pattern = '\\{}.*?\\{}'.format(*mod_ident)
    modpep = modpep.replace('_', '')
    mod_len = 0
    mod = ''
    for _ in re.finditer(re_find_pattern, modpep):
        _start, _end = _.span()
        mod += '{},{};'.format(_start - mod_len, _.group().strip(''.join(mod_ident)))
        mod_len += _end - _start
    stripped_pep = re.sub(re_sub_pattern, '', modpep)
    return stripped_pep, mod


def add_mod(pep, mod, mod_processor):
    """
    mod_process is the ModOperation class
    """
    if mod:
        if isinstance(mod, str):
            mod = str_mod_to_list(mod)
        mod = sorted(mod, key=lambda x: x[0])
        mod_pep_list = []
        prev_site_num = 0
        for mod_site, mod_name in mod:
            mod_pep_list.append(pep[prev_site_num: mod_site])
            if mod_site != 0:
                mod_aa = mod_pep_list[-1][-1]
            else:
                mod_aa = pep[0]
            mod_pep_list.append(mod_processor(mod=mod_name, aa=mod_aa))
            prev_site_num = mod_site
        mod_pep_list.append(pep[prev_site_num:])
        mod_pep = ''.join(mod_pep_list)
    else:
        mod_pep = pep
    mod_pep = f'_{mod_pep}_'
    return mod_pep


def fasta_title(title: str, title_type='uniprot'):
    title = title.lstrip('>')

    if '|' in title:
        ident = title.split('|')[1]
    else:
        ident = title.split(' ')[0]
    return ident


def join_seqtext_in_fasta(seq_text):
    return ''.join(seq_text.split('\n'))


def join_seqlines_in_fasta(seq_lines):
    return ''.join([_.strip('\n') for _ in seq_lines])




PRM_Header = [
    'Mass [m/z]',  # MS1 m/z
    'Formula [M]',
    'Formula type',
    'Species',
    'CS [z]',  # Integer
    'Polarity',  # "Positive"
    'Start [min]',
    'End [min]',
    '(N)CE',
    '(N)CE type',
    'MSX ID',
    'Comment',
]



import numpy as np
import tensorflow as tf
from tensorflow.python import pywrap_tensorflow


def count_param_num(ckpt_path, show_single_varible=False):
    with tf.Graph().as_default():
        reader = pywrap_tensorflow.NewCheckpointReader(ckpt_path)
        all_variables = reader.get_variable_to_shape_map()

        # print(all_variables)
        total_parameters = 0
        for key in all_variables:

            shape = np.shape(reader.get_tensor(key))
            shape = list(shape)
            # print(reader.get_tensor(key))
            # print('dims', len(shape))
            variable_parameters = 1
            for dim in shape:
                # print(dim)
                variable_parameters *= dim
            total_parameters += variable_parameters

            if show_single_varible:
                print('=' * 30)
                print(key)
                print('Shape:', shape)
                print('Parameters of current variable:', variable_parameters)
        print('=' * 30)
        print(f'{total_parameters} parameters ({total_parameters / 1000:.3f}K) ({total_parameters / 1000000:.3f}M)')


def count_flops(graph):
    flops = tf.profiler.profile(graph, options=tf.profiler.ProfileOptionBuilder.float_operation())
    print('FLOPs: {}'.format(flops.total_float_ops))



try:
    import cv2
except ModuleNotFoundError:
    pass


def read_img_to_rgb(img_path):
    try:
        cv2
    except NameError:
        print('OpenCV is not installed. If you want to use this function, please install via pip `pip install opencv-python`')
    fig = cv2.imread(img_path, cv2.IMREAD_LOAD_GDAL)
    bgr_ = cv2.split(fig)
    trans_fig = cv2.merge(bgr_[::-1]) if len(bgr_) == 3 else cv2.merge(bgr_[:3][::-1] + bgr_[-1])
    return trans_fig



import numpy as np
import pandas as pd

from mskit.rapid_kit import substring_finder
from .sn_utils import sn_modpep_to_intseq


def init_phos_analysis(df,
                       protein_col='PG.ProteinGroups',
                       modpep_col='EG.ModifiedPeptide',
                       prec_charge_col='FG.Charge',
                       ):
    df['FirstProtein'] = df[protein_col].apply(lambda x: x.split(';')[0])
    df['Precursor'] = df[modpep_col] + '.' + df[prec_charge_col].astype(int).astype(str)
    df['IntPep'] = df[modpep_col].apply(sn_modpep_to_intseq)
    df['IntPrec'] = df['IntPep'] + '.' + df[prec_charge_col].astype(str)

    df['PhosNum'] = df[modpep_col].apply(lambda x: x.count('[Phospho (STY)]'))
    df['IsPhospep'] = df['PhosNum'].apply(lambda x: True if x > 0 else False)
    df['IsMod'] = df[modpep_col].apply(lambda x: True if '[' in x else False)
    df['DropPhosPep'] = df[modpep_col].apply(lambda x: x.replace('[Phospho (STY)]', ''))
    return df


def add_phossite_in_pep(x, modpep_col='EG.ModifiedPeptide',
                        prob_thres=None,
                        prob_col='EG.PTMProbabilities [Phospho (STY)]',
                        modpos_col='EG.PTMPositions [Phospho (STY)]',
                        required_mod='[Phospho (STY)]'):
    if not x['IsPhospep']:
        return np.nan

    strippep, modpos, mods = substring_finder(x[modpep_col].replace('_', ''))
    required_mod_pos = [pos for pos, mod in zip(modpos, mods) if mod == required_mod]

    if prob_thres:
        prob_list = list(map(float, x[prob_col].split(';')))
        peppos_list = list(map(int, x[modpos_col].split(';')))
        thres_filter_pos = [peppos
                            for prob, peppos
                            in zip(prob_list, peppos_list)
                            if peppos in required_mod_pos and prob >= prob_thres]
    else:
        thres_filter_pos = required_mod_pos

    if thres_filter_pos:
        if len(thres_filter_pos) == len(required_mod_pos):
            return ';'.join(list(map(str, thres_filter_pos)))
        else:
            return 'SiteFilter'
    else:
        return 'PepFilter'


def add_phossite_in_prot(x, ):
    if pd.isna(x['PepPhosPos']):
        return np.nan

    peppos_in_prot = str(x['PEP.PeptidePosition']).split(';')
    prot_mod_site = ''
    pep_mod_sites = x['PepPhosPos'].split(';')

    for protein, peppos in zip(x['PG.ProteinGroups'].split(';'), peppos_in_prot):
        if peppos == '':
            prot_mod_site += ';'
            continue
        for one_pep_pos in peppos.split(','):
            for one_mod_site in pep_mod_sites:
                prot_mod_site += f'{int(one_pep_pos) + int(one_mod_site) - 1},'
            prot_mod_site = prot_mod_site.strip(',')
            prot_mod_site += '/'
        prot_mod_site = prot_mod_site.strip('/')
        prot_mod_site += ';'
    prot_mod_site = prot_mod_site[:-1]

    return prot_mod_site



import os
import re

import pandas as pd

from mskit import rapid_kit
from ._deeprt_constant import MOD


def newrt_format_to_deeprt(rt_df: pd.DataFrame, pep_col='IntPep'):
    rt_df = rt_df[rt_df[pep_col].apply(lambda x: True if x[0] != '*' else False)]
    rt_df['IntPep'] = rt_df['IntPep'].apply(lambda x: x[1:] if x[0] == '@' else x)
    return rt_df


def parse_deeprt_log(log_path):
    with open(log_path, 'r') as f:
        log_content = f.readlines()

    param_num = log_content[0].split(':')[1]
    epoch_num = (len(log_content) - 2) // 3
    log_info = []
    for epoch_idx in range(1, epoch_num + 1):
        rows = log_content[(epoch_idx - 1) * 3 + 1: epoch_idx * 3 + 1]
        train_loss = re.findall(r'Training Loss: (\d+?\.\d+) ', rows[0])[0]
        test_loss = re.findall(r'Testing Loss: (\d+?\.\d+) ', rows[1])[0]
        data_num, pearson, spearman = re.findall(r'Corr on (\d+) testing samples: (\d+?\.\d+) \| (\d+?\.\d+)', rows[2])[0]
        log_info.append((epoch_idx, train_loss, test_loss, pearson, spearman, data_num))
    log_df = pd.DataFrame(log_info, columns=['Epoch', 'TrainLoss', 'TestLoss', 'TestPearson', 'TestSpearman', 'DataNum'])
    log_df = log_df.astype({'Epoch': int, 'DataNum': int, 'TrainLoss': float, 'TestLoss': float, 'TestPearson': float, 'TestSpearman': float})
    return log_df, param_num


def deeprt_input(out_path, pep_list, mod_trans=True):
    with open(out_path, "w") as f:
        f.write("sequence\tRT\n")
        for pep in pep_list:
            if mod_trans:
                pep = trans_sn_mod(pep)
                if not pep:
                    continue
            f.write(pep + "\t" + "0.0000" + "\n")


def deeprt_trainset(output_path, modpep_rt_list, mod_trans=True):
    with open(output_path, 'w') as out_file:
        out_file.write('sequence\tRT\n')
        for each_modpep_rt in modpep_rt_list:
            each_modpep = each_modpep_rt[0]
            each_rt = each_modpep_rt[1]
            if mod_trans:
                each_modpep = trans_sn_mod(each_modpep)
                if not each_modpep:
                    continue
            out_file.write('{seq}\t{rt}\n'.format(seq=each_modpep, rt=each_rt))


def read_deeprt_pred(pred_file, method='pandas', mod_trans=True):
    if method == 'pandas':
        pred_output = pd.read_csv(os.path.abspath(pred_file), sep='\t')
        seq = pred_output['seq'].to_list()
        pred_rt = pred_output['predicted'].to_list()
    else:
        with open(pred_file, 'r') as _pred:
            pred_title = _pred.readline()
            seq = []
            pred_rt = []
            for each_line in _pred:
                split_line = each_line.strip('\n').split('\t')
                seq.append(split_line[0])
                pred_rt.append(split_line[2])
    if mod_trans:
        modpep_list = []
        for modpep in seq:
            for sn_mod, int_mod in MOD.items():
                modpep = modpep.replace(int_mod, sn_mod)
            modpep = f'_{modpep}_'
            modpep_list.append(modpep)
    else:
        modpep_list = seq
    return dict(zip(modpep_list, pred_rt))


def get_coincide_rtdata(rt_dict_1, rt_dict_2):
    modpep, rt_list_1, rt_list_2 = rapid_kit.get_coincide_data(rt_dict_1, rt_dict_2)
    return modpep, rt_list_1, rt_list_2


def trans_sn_mod(modpep):
    for sn_mod, int_mod in MOD.items():
        modpep = modpep.replace(sn_mod, int_mod)
        if '[' not in modpep:
            break
    if '[' in modpep:
        return None
    modpep = modpep.strip('_')
    return modpep



"""
ProteomeXchange
    Pride

"""


class ProteomeXchange:
    PXD_Dataset = 'http://proteomecentral.proteomexchange.org/cgi/GetDataset?ID={}'

    class Pride:
        """
        https://pride-archive.github.io/PrideAPIDocs/hyperlinks.html

        ListQueryUrl:
            Url + Project ID (PXD...)
                e.g. https://www.ebi.ac.uk/pride/ws/archive/file/list/project/PXD004732
            This will get a json file
        Example FTP file address
            ftp://ftp.pride.ebi.ac.uk/pride/data/archive/2018/06/PXD009449/MaxQuant_1.5.3.30.zip
        Example https file address
            https://ftp.pride.ebi.ac.uk/pride/data/archive/2019/07/PXD008355/txt.zip

        """
        ListQueryUrl = 'https://www.ebi.ac.uk/pride/ws/archive/file/list/project/'
        FTP_Url = 'ftp://ftp.pride.ebi.ac.uk'
        ASP_Url = 'era-fasp@fasp.pride.ebi.ac.uk:'
        FTP_IP = '193.62.197.74'
        HTTPS_Url = 'https://ftp.pride.ebi.ac.uk'

        PRIDE_StoragePrefix = '/pride/data/archive/'


    class Massive:
        FTP_Url = 'ftp://massive.ucsd.edu'
        FTP_IP = '132.249.211.16'

    class JPOST:
        FTP_Url = 'ftp://ftp.biosciencedbc.jp'
        StoragePrefix = '/archive/jpostrepos/'
        FTP_IP = '160.74.86.87'

        ProjectInfo = 'https://repository.jpostdb.org/entry/{}'  # JPST000859
        ProjectXML = 'https://repository.jpostdb.org/xml/{}.{}.xml'  # JPST000859, 3
        SingleFileUrl = 'https://repository.jpostdb.org/data/{}.{}/{}'  # JPST000859, 3, FileName


class DGI:
    """
    Drug Gene Interaction: http://www.dgidb.org/
    """
    DGIInteractionURL = r'http://dgidb.org/api/v2/interactions.json?genes={}'


class KEGG:
    KEGGDrugDatabaseFind = r'http://rest.kegg.jp/find/drug/{}'  # Fill in drug name which is connected with plus sign
    KEGGGetURL = r'http://rest.kegg.jp/get/{}'  # Fill in drug D number


class STRING:
    STRING_API_URL = "https://string-db.org/api"
    Methods = ['interaction_partners']
    OutputFormat = []


class Uniprot:
    MainUrl = 'https://www.uniprot.org/'



import os
import re
import typing

import numpy as np

from mskit import rapid_kit
from .ted import TED

__all__ = [
    'read_fasta',
    'write_fasta',
    'FastaFile'
]


def read_fasta(
        fasta_file,
        sep: typing.Union[None, str] = '|',
        ident_idx: int = 1,
        ident_process_func: typing.Union[None, typing.Callable] = None,
        open_mode: str = 'r',
        skip_row=None,
        ignore_blank=False
):
    """
    Use `sep=None` to skip parsing title

    """
    fasta_dict = dict()
    seq_list = []
    with open(fasta_file, open_mode, ) as f:
        if isinstance(skip_row, int):
            [f.readline() for _ in range(skip_row)]
        for row in f:
            if open_mode == 'rb':
                row = row.decode()
            if row.startswith('>'):
                if seq_list:
                    fasta_dict[acc] = ''.join(seq_list)

                if ident_process_func is not None:
                    acc = ident_process_func(acc)
                else:
                    if sep is None:
                        acc = row.strip('\n')
                    else:
                        acc = row.strip('\n').split(sep)[ident_idx]

                seq_list = []
            elif not row or row == '\n':
                if ignore_blank:
                    continue
                else:
                    raise ValueError('Blank line in target FASTA file. '
                                     'Check completeness of FASTA file, or use `ignore_blank=True` to ignore this error.')
            else:
                seq_list.append(row.strip('\n'))
        if seq_list:
            fasta_dict[acc] = ''.join(seq_list)
    return fasta_dict


seq_dict_from_fasta = read_fasta


def write_fasta(
        fasta: dict,
        file_path: str,
        seq_line_max_char: int = None
):
    """
    line_max_char:
        None: Write one seq to one line
        80: Write one seq to multilines to keep the numebr of char in one line == 80
        other integer: Keep number of char equal to the customed number
    """
    with open(file_path, 'w') as f:
        if seq_line_max_char is not None:
            for title, seq in fasta.items():
                f.write(title + '\n')
                f.write(
                    ''.join([
                        seq[_ * seq_line_max_char: (_ + 1) * seq_line_max_char] + '\n'
                        for _ in range(int(np.ceil(len(seq) / seq_line_max_char)))
                    ])
                )
        else:
            for title, seq in fasta.items():
                f.write(title + '\n')
                f.write(seq + '\n')


class FastaFile(object):
    """
    TODO 传入path，或content，或handle，增加skiprow和commend ident
    TODO 两个 fasta parser 合并
    """

    def __init__(
            self,
            fasta_file_path,
            open_mode='r',
            title_parse_rule: typing.Union[str, typing.Callable] = 'uniprot',
            preprocess=True,
            nothing_when_init=False
    ):
        """
        :param fasta_file_path:
        :param title_parse_rule: 'uniprot' will get the second string for title split by '|', and others will be the first string split by get_one_prefix_result blank,
        while maybe other formats of fasta title is needed later
        """

        self.fasta_file_path = fasta_file_path
        self.raw_entry_seq_map = dict()
        self.title_parse_rule = title_parse_rule

        self.id_parse_func = None
        self.comment_parse_func = None

        self.prot_acc_dict = dict()

        self._protein_info = dict()  # The description information of each protein in the fasta file
        self._protein_to_seq = dict()  # The whole sequence of each protein (No digestion)
        self._seq_to_protein = dict()  # Digested peptide to protein. The protein may be str if one else list.
        self._seq_list = []  # Digested peptides of all protein sequence in the fasta file

        if preprocess:
            self.load_fasta()

    def get_fasta_file_path(self):
        return self.fasta_file_path

    def set_fasta_file_path(self, fasta_file_path):
        if os.path.exists(fasta_file_path):
            self.fasta_file_path = os.path.abspath(fasta_file_path)
        else:
            print('Incorrect FASTA file path')
            raise FileNotFoundError(f'The FASTA file is not existed: {fasta_file_path}')

    fasta_file_path = property(get_fasta_file_path, set_fasta_file_path, doc='''''')

    def load_fasta(self, open_mode='r', method='re'):
        if method == 're':
            with open(self.fasta_file_path, 'r') as f:
                raw_content = f.read()
            title_seq_list = re.findall('(>.+?\\n)([^>]+\\n?)', raw_content)
            title_seq_list = [(title.strip('\n'), seq.replace('\n', '')) for title, seq in title_seq_list]
            self.raw_entry_seq_map = dict(title_seq_list)
        else:
            self.raw_entry_seq_map = read_fasta(
                self.fasta_file_path,
                sep=None,
                ident_idx=-1,
                ident_process_func=None,
                open_mode=open_mode,
                skip_row=None,
                ignore_blank=False
            )

    def get_title_parse_rule(self):
        return self.title_parse_rule

    def set_title_parse_rule(
            self,
            parse_rule=None,
            id_parse_func=None,
            comment_parse_func=None
    ):
        print('enter set parse func')
        id_parse_rules = {'uniprot': lambda x: re.findall('^>(.+?)\|(.+?)\|(.+?)$', x)}
        comment_parse_rules = {'uniprot': lambda x: None}
        if isinstance(parse_rule, str):
            try:
                self.id_parse_func = id_parse_rules[parse_rule.lower()]  # TODO 设置一个 parse rule dict
            except KeyError:
                raise KeyError(f'Not {parse_rule} Found in The Predefined rule list')
        else:
            pass
        self.title_parse_rule = parse_rule

        if id_parse_func:
            self.id_parse_func = id_parse_func
        if comment_parse_func:
            self.comment_parse_func = comment_parse_func

    title_parse_rule = property(get_title_parse_rule, set_title_parse_rule, doc='''''')

    def __call__(self, *args, **kwargs):
        pass

    def __iter__(self):
        return iter(self.raw_entry_seq_map.items())

    def __getitem__(self, item):
        return self.prot_acc_dict[item]

    def __setitem__(self, key, value):
        self.raw_entry_seq_map[key] = value

    def __add__(self, other):  # 只保留唯一 title
        pass

    def get_all_title(self):
        pass

    def add_new_seqs(self, new_seq_dict, id_conflict=None):
        """
        id_conflict:
            new: Keep new protein seq
            origin: Keep original protein seq
            go_on: Add as PROTEIN-2
            consistent: Add as PROTEIN-2 and rename the original key to PROTEIN-1
        """
        self.raw_entry_seq_map.update(new_seq_dict)

    def to_file(self, file_path, seq_line=None):
        """
        seq_line:
            None: Write one seq to one line
            80: Write one seq to multilines to keep the numebr of char in one line == 80
            other integer: Keep number of char equal to the customed number
        """
        with open(file_path, 'w') as f:
            if seq_line:
                for title, seq in self.raw_entry_seq_map.items():
                    f.write(title + '\n')
                    f.write(''.join([seq[_ * seq_line: (_ + 1) * seq_line] + '\n' for _ in
                                     range(int(np.ceil(len(seq) / seq_line)))]))
            else:
                for title, seq in self.raw_entry_seq_map:
                    f.write(title + '\n')
                    f.write(seq + '\n')

    save = to_file

    def one_protein_generator(self):
        """
        Generate title and sequence of each protein in fasta file
        """
        seq_title = ''
        seq_list = []
        with open(self.fasta_path, 'r') as fasta_handle:
            for _line in fasta_handle:
                if not _line:
                    print('Blank line existed in fasta file')  # TODO 记录 blank line 的行号
                    continue
                if _line.startswith('>'):
                    if seq_title and seq_list:
                        yield seq_title, ''.join(seq_list)
                    seq_title = _line.strip('\n')
                    seq_list = []
                else:
                    seq_list.append(_line.strip('\n'))
            if seq_title and seq_list:
                yield seq_title, ''.join(seq_list)

    def protein2seq(self, protein_info=False):
        if not self._protein_to_seq:
            for _title, _seq in self.one_protein_generator():
                protein_ident = rapid_kit.fasta_title(_title, self.title_parse_rule)
                self._protein_to_seq[protein_ident] = _seq
                if protein_info:
                    self._protein_info[protein_ident] = _title
        return self._protein_to_seq

    def seq2protein(
            self,
            miss_cleavage=(0, 1, 2),
            min_len=7,
            max_len=33) -> dict:

        if not self._seq_to_protein:
            if not self._protein_to_seq:
                self.protein2seq()

            ted = TED(miss_cleavage=miss_cleavage,
                      min_len=min_len,
                      max_len=max_len,
                      enzyme='Trypsin',
                      return_type='seq')
            for protein_acc, seq in self._protein_to_seq.items():
                compliant_seq = ted(seq)
                for _each_seq in compliant_seq:
                    self._seq_list.append(_each_seq)
                    if _each_seq not in self._seq_to_protein:
                        self._seq_to_protein[_each_seq] = protein_acc
                    else:
                        if isinstance(self._seq_to_protein[_each_seq], str):
                            self._seq_to_protein[_each_seq] = [
                                self._seq_to_protein[_each_seq], protein_acc]
                        elif isinstance(self._seq_to_protein[_each_seq], list):
                            self._seq_to_protein[_each_seq].append(
                                protein_acc)
        return self._seq_to_protein

    def get_total_seqlist(
            self,
            miss_cleavage=(0, 1, 2),
            min_len=7,
            max_len=33):

        if not self._seq_list:
            self.seq2protein(miss_cleavage=miss_cleavage,
                             min_len=min_len,
                             max_len=max_len)
        self._seq_list = rapid_kit.drop_list_duplicates(self._seq_list)
        return self._seq_list

    def merge(self):
        pass


def merge_fasta(
        *fasta: typing.Union[FastaFile, dict],
        duplicated_title: str = 'keep_first',
        duplicated_seq: str = 'keep_all',
):
    """
    Merge input FASTAs to one.
    :param fasta:
        Receives multi input FASTAs for `FastaFile` object or dict
        If inputs are all `FastaFile` object, this function will return
        the first `FastaFile` object (same id) with combined `protein: seq` pairs
        If any one input is `dict`, this function will return a dict with combined `protein: seq` pairs
    :param duplicated_title: one of "keep_first", "drop_all", "continous_anno"
    :param duplicated_seq: one of "keep_all", "drop_all". This will be checked after title checking
    """

    pass


class FastaWriter(object):
    def __init__(self, path):
        # TODO Check defined path
        self.fasta_path = path
        self.fasta_file_stream = None

    def __enter__(self):
        self.fasta_file_stream = open(self.fasta_path, 'r')

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.fasta_file_stream.close()


class _FastaParser(FastaFile, ):
    def __init__(self, fasta_type='protein'):
        if fasta_type.lower() == 'protein':
            super(FastaFile, self).__init__()
        elif fasta_type.lower() == 'base' or fasta_type.lower() == 'nucleic acid':
            pass



import numpy as np


def normalize_intensity(inten, max_num=None, exclude_low_inten=None, no_drop=False):
    """
    The max_num here is the maximum number of intensity, and it will be 1 if is not defined
    输入 list 时可以返回和输入相同的长度，低于 low inten 的部分填充 False
    """
    if max_num is None:
        max_num = 1
    if exclude_low_inten is None:
        exclude_low_inten = 0
    if isinstance(inten, dict):
        max_intensity = max(inten.values())
        scale = max_num / max_intensity
        inten_dict = dict()
        for each_fragment, frag_inten in inten.items():
            scaled_inten = frag_inten * scale
            if scaled_inten > exclude_low_inten:
                inten_dict[each_fragment] = scaled_inten
        return inten_dict
    elif isinstance(inten, (list, tuple, np.ndarray)):
        max_intensity = max(inten)
        scale = max_num / max_intensity
        scaled_inten_list = [_ * scale for _ in inten]
        return [_ for _ in scaled_inten_list if _ > exclude_low_inten]
    else:
        raise


def keep_top_n_inten(inten, top_n=25):
    """
    可以作为 low inten filter 加入 normalize_intensity
    """
    if len(inten) <= top_n:
        return inten
    if isinstance(inten, dict):
        top_n_inten = sorted(inten.values(), reverse=True)[top_n]
        return {frag: inten_value for frag, inten_value in inten.items() if inten_value > top_n_inten}
    elif isinstance(inten, list):
        top_n_inten = sorted(inten, reverse=True)[top_n]
        return [inten_value for inten_value in inten if inten_value > top_n_inten]
    else:
        raise



import os


def list_dir_with_identification(
        dirname,
        identification=None,
        position='end',
        regex=False,
        full_path=False
):
    dir_content_list = os.listdir(dirname)
    if identification:
        if position == 'end':
            dir_content_list = [
                _ for _ in dir_content_list if _.endswith(identification)]
        elif position == 'in':
            dir_content_list = [
                _ for _ in dir_content_list if identification in _]
        else:
            raise NameError('parameter position is illegal')
    if not full_path:
        return dir_content_list
    else:
        return [os.path.join(dirname, _) for _ in dir_content_list]


def get_workspace(level=0):
    curr_dir = os.path.abspath('.')
    work_dir = curr_dir
    for i in range(level):
        work_dir = os.path.dirname(work_dir)
    return work_dir



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from . import gradients


class PreColor:
    red_gradient = np.array(gradients.gradient_hsl((255, 218, 185), (35, 0, 0), 200)) / 256
    green_gradient = np.array(gradients.gradient_hsl((240, 255, 240), (0, 35, 0), 200)) / 256
    blue_gradient = np.array(gradients.gradient_hsl((240, 255, 255), (0, 0, 35), 200)) / 256

    yellow_gradient = np.array(gradients.gradient_hsl((255, 250, 205), (255, 255, 0), 200)) / 256
    violet_gradient = np.array(gradients.gradient_hsl((230, 230, 250), (255, 0, 255), 200)) / 256
    cyan_gradient = np.array(gradients.gradient_hsl((230, 230, 250), (0, 255, 255), 200)) / 256
    purple_gradient = sns.light_palette(np.array((50, 0, 60)) / 256, input="rgb", n_colors=200)

    rose_gradient = np.array(gradients.gradient_hsl((250, 230, 230), (255, 1, 1), 200)) / 256
    grass_gradient = np.array(gradients.gradient_hsl((230, 250, 230), (0, 255, 0), 200)) / 256
    blue_violet_gradient = np.array(gradients.gradient_hsl((230, 230, 250), (0, 0, 255), 200)) / 256

    pink_gradient = None
    brown_gradient = None

    grey_gradient = np.array(gradients.gradient_rgb((255, 255, 255), (0, 0, 0), 200)) / 256


PreColorDict = {
    'Red': PreColor.red_gradient,
    'Green': PreColor.green_gradient,
    'Blue': PreColor.blue_gradient,
    'Yellow': PreColor.yellow_gradient,
    'Violet': PreColor.violet_gradient,
    'Purple': PreColor.purple_gradient,
    'Cyan': PreColor.cyan_gradient,
    'Rose': PreColor.rose_gradient,
    'Grass': PreColor.grass_gradient,
    'Blue-violet': PreColor.blue_violet_gradient,
    'Grey': PreColor.grey_gradient,
}


class FavorGradColor:
    GreyBlueRed = gradients.gradient_hsl(
        (230, 235, 240), (30, 40, 60), 150, value_scale=256)[::-1] + gradients.gradient_hsl(
        (245, 250, 255), (230, 235, 240), 35, value_scale=256)[::-1] + gradients.gradient_hsl(
        (255, 250, 245), (240, 230, 220), 35, value_scale=256) + gradients.gradient_hsl(
        (240, 230, 220), (100, 10, 10), 150, value_scale=256)
    BlueRed = sns.color_palette('RdBu_r', n_colors=400)
    DardRed = sns.mpl_palette("Reds_d", 400)[::-1]


def draw_grad_precolors():
    c_num = len(PreColorDict)
    f, axes = plt.subplots(1, c_num, figsize=(c_num, 1), dpi=150)

    for ax_index, (c_name, color) in enumerate(PreColorDict.items()):
        ax = axes[ax_index]
        ax.grid(False)

        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)

        for _ in ['bottom', 'top', 'left', 'right']:
            ax.spines[_].set_visible(False)

        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)

        n = 200
        _color = color[:n]

        for i, c in enumerate(_color):
            ax.plot((0, 1), (i / n, i / n), color=c)

        ax.annotate(0, (1, 0),
                    textcoords='offset points', xytext=(0.5, 0),
                    va='center', ha='left', fontsize=9, rotation=0)
        ax.annotate(1, (1, 1),
                    textcoords='offset points', xytext=(0, 0),
                    va='center', ha='left', fontsize=9, rotation=0)

        ax.annotate(c_name, (0.5, 1),
                    textcoords='offset points', xytext=(0, 5),
                    va='bottom', ha='center', fontsize=9, rotation=0)



import re

from mskit.inherited_builtins import NonOverwriteDict


def process_name_line(line_content):
    split_line = line_content.strip('\n').split(' ')
    stripped_pep, charge = split_line[1].split('/')
    return stripped_pep, charge


def process_comment_line(line_content):
    split_line = line_content.strip('\n').split(' ')
    mods = split_line[2]
    irt = split_line[4].strip('iRT=')
    mod_content_split = mods.split('/')
    mod_num = mod_content_split[0].strip('Mods=')
    if mod_num == '0':
        mod_list = None
    else:
        mod_list = [(int(_.split(',')[0]), _.split(',')[1], _.split(',')[2]) for _ in set(mod_content_split[1:])]
        mod_list = sorted(mod_list, key=lambda x: x[0])
    return mod_list, irt


def peptide_add_mod(stripped_peptide, mod_list):
    mod_pep = ''
    previous_mod_site = 0
    for each_mod in mod_list:
        mod_site = each_mod[0] + 1
        mod_aa = each_mod[1]
        mod_type = each_mod[2]
        mod_pep += stripped_peptide[previous_mod_site: mod_site] + f'[{mod_type} ({mod_aa})]'
        previous_mod_site = mod_site
    mod_pep += stripped_peptide[previous_mod_site:]
    return mod_pep


def process_fragment_line(line_content):
    split_line = line_content.strip('\n').split('\t')
    mz = split_line[0]
    intensity = float(split_line[1])
    fragment_description = split_line[2].replace('"', '').split('/')
    fragment_identifier = fragment_description[0]
    fragment_type_num = re.findall('([by]\d+)', fragment_identifier)[0]
    fragment_charge = re.findall('\^(\d)', fragment_identifier)
    fragment_losstype = re.findall('-(.+?)(\^|$)', fragment_identifier)
    fragment_name = fragment_type_num
    if fragment_charge:
        fragment_name += '+{}'.format(fragment_charge[0])
    else:
        fragment_name += '+1'
    if fragment_losstype:
        fragment_name += '-{}'.format(fragment_losstype[0][0])
    else:
        fragment_name += '-noloss'
    return fragment_name, intensity


def read_synthetic_data(synthetic_file):
    with open(synthetic_file, 'r') as f:
        synthetic_data = dict()
        fragment_info = NonOverwriteDict()
        for line_series, each_line in enumerate(f):
            if each_line.startswith(' '):
                continue
            elif each_line == '\n':
                if mod_list:
                    precursor = '_{}_.{}'.format(peptide_add_mod(stripped_pep, mod_list), charge)
                else:
                    precursor = '_{}_.{}'.format(stripped_pep, charge)
                synthetic_data[precursor] = {'iRT': irt, 'Fragment': fragment_info, 'Precursor': precursor,
                                             'Charge': charge, 'StrippedPeptide': stripped_pep}
                fragment_info = NonOverwriteDict()
            elif each_line.startswith('Name'):
                stripped_pep, charge = process_name_line(each_line)
            elif each_line.startswith('MW'):
                continue
            elif each_line.startswith('Comment'):
                mod_list, irt = process_comment_line(each_line)
            elif each_line.startswith('Num peaks'):
                continue
            elif each_line[0].isdigit():
                fragment_name, intensity = process_fragment_line(each_line)
                fragment_info[fragment_name] = intensity
            else:
                print('Error in line {}'.format(line_series + 1))
    return synthetic_data



"""
1.
This stores many different modification display methods, and all the modification will be got from here.

Shorthand:
Spectronaut -> SN

"""


class SNMod(object):
    SNModToUnimod_13 = {
        ''
    }

    def __init__(self, sn_version=14):
        if sn_version >= 13:
            self.sn_mod_to_unimod = self.SNModToUnimod_13
        else:
            pass

    def sn_modpep_to_unimod_pep(self):
        pass


class LossType(object):
    SN_to_Readable = {
        'noloss': 'Noloss',

        'H2O': '1,H2O',
        'NH3': '1,NH3',

        'H3PO4': '1,H3PO4',
        '1(+H2+O)1(+H3+O4+P)': '1,H2O;1,H3PO4',
        '1(+H3+N)1(+H3+O4+P)': '1,NH3;1,H3PO4',

        '2(+H3+O4+P)': '2,H3PO4',
        '1(+H2+O)2(+H3+O4+P)': '1,H2O;2,H3PO4',
        '1(+H3+N)2(+H3+O4+P)': '1,NH3;2,H3PO4',

        '3(+H3+O4+P)': '3,H3PO4',
        '1(+H2+O)3(+H3+O4+P)': '1,H2O;3,H3PO4',
        '1(+H3+N)3(+H3+O4+P)': '1,NH3;3,H3PO4',

        '4(+H3+O4+P)': '4,H3PO4',
        '1(+H2+O)4(+H3+O4+P)': '1,H2O;4,H3PO4',
        '1(+H3+N)4(+H3+O4+P)': '1,NH3;4,H3PO4',

        '5(+H3+O4+P)': '5,H3PO4',
        '1(+H2+O)5(+H3+O4+P)': '1,H2O;5,H3PO4',
        '1(+H3+N)5(+H3+O4+P)': '1,NH3;5,H3PO4',

    }

    Readable_to_SN = {v: k for k, v in SN_to_Readable.items()}
    PreOrder = list(SN_to_Readable.keys())

    SN_to_Readable_v13 = SN_to_Readable.copy()
    SN_to_Readable_v13['1(+H9+N+O8+P2)'] = '1,NH3;2,H3PO4',  # This will occur in SN 13


class BasicModInfo(object):
    ModDict_new = {'C': 'C[Carbamidomethyl (C)]',
                   'M': 'M[Oxidation (M)]'}
    ModDict_old = {'C': 'C[Carbamidomethyl]',
                   'M': 'M[Oxidation]'}

    ModConvert_new = {}
    ModConvert_old = {}

    DeepRTIntModDict_new = {
        'C[Carbamidomethyl (C)]': 'C',
        'M[Oxidation (M)]': '1',
        'S': '2',
        'T': '3',
        'Y': '4'}
    DeepRTIntModDict_old = {
        'C[Carbamidomethyl]': 'C',
        'M[Oxidation]': '1',
        'S': '2',
        'T': '3',
        'Y': '4'}
    # IntModDict = {'M[+16]': '1', 'S[+80]': '2', 'T[+80]': '3', 'Y[+80]': '4'}

    pDeepModType = {'C': 'Carbamidomethyl[C]', 'M': 'Oxidation[M]'}


class ModType(BasicModInfo):
    """
    Spectronaut version 12 has get_one_prefix_result different modification display type.
    The default version is set to 12, which uses the new modification display method.
    The version should be set in each main functions but not the functions that are used frequently.
    """

    def __init__(self, spectronaut_version=12):
        self._spectronaut_version = spectronaut_version
        self.ModDict = self.ModDict_new
        self.DeepRTIntModDict = self.DeepRTIntModDict_new

        self.pDeepMod2SNMod_new = dict(
            [(self.pDeepModType[aa], self.ModDict_new[aa][1:]) for aa in self.pDeepModType])
        self.pDeepMod2SNMod_old = dict(
            [(self.pDeepModType[aa], self.ModDict_old[aa][1:]) for aa in self.pDeepModType])
        self.pDeepMod2SNMod = self.pDeepMod2SNMod_new

    def set_spectronaut_version(self, version):
        self._spectronaut_version = version
        if self._spectronaut_version >= 12:
            self.ModDict = self.ModDict_new
            self.DeepRTIntModDict = self.DeepRTIntModDict_new
            self.pDeepMod2SNMod = self.pDeepMod2SNMod_new
        else:
            self.ModDict = self.ModDict_old
            self.DeepRTIntModDict = self.DeepRTIntModDict_old
            self.pDeepMod2SNMod = self.pDeepMod2SNMod_old

    def get_spectronaut_int_version(self):
        return self._spectronaut_version

    def get_spectronaut_str_version(self):
        if self._spectronaut_version >= 12:
            return 'new'
        else:
            return 'old'

    @staticmethod
    def get_mod_dict(ver='new'):
        if isinstance(ver, str):
            if ver == 'new':
                return ModType.ModDict_new
            elif ver == 'old':
                return ModType.ModDict_old
            else:
                raise NameError('Choose mod version from \'new\' and \'old\'')
        elif isinstance(ver, int):
            if ver >= 12:
                return ModType.ModDict_new
            else:
                return ModType.ModDict_old


class SNLibraryTitle(object):
    LibraryMainCol = [
        'PrecursorCharge',
        'ModifiedPeptide',
        'StrippedPeptide',
        'iRT',
        'LabeledPeptide',
        'PrecursorMz',
        'FragmentLossType',
        'FragmentNumber',
        'FragmentType',
        'FragmentCharge',
        'FragmentMz',
        'RelativeIntensity',
        'ProteinGroups']
    LibraryMainColPGOut = [
        'PrecursorCharge',
        'ModifiedPeptide',
        'StrippedPeptide',
        'iRT',
        'LabeledPeptide',
        'PrecursorMz',
        'FragmentLossType',
        'FragmentNumber',
        'FragmentType',
        'FragmentCharge',
        'FragmentMz',
        'RelativeIntensity']



from matplotlib import pyplot as plt
from matplotlib.ticker import FuncFormatter


def remove_xy_ticks(ax=None):
    if ax is None:
        ax = plt.gca()
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)


def remove_target_spines(spine_pos, ax=None):
    if ax is None:
        ax = plt.gca()
    if isinstance(spine_pos, str):
        spine_pos = [spine_pos]
    for _ in spine_pos:
        ax.spines[_].set_visible(False)


def remove_all_spines(ax=None):
    if ax is None:
        ax = plt.gca()
    for _ in ['top', 'right', 'bottom', 'left']:
        ax.spines[_].set_visible(False)


def remove_all_bone_info(ax=None):
    if ax is None:
        ax = plt.gca()
    remove_xy_ticks(ax)
    remove_all_spines(ax)


def set_thousand_separate(ax=None, axis=('x', 'y')):
    if ax is None:
        ax = plt.gca()
    if not isinstance(axis, tuple):
        axis = tuple(axis)
    for a in axis:
        if a == 'x':
            ax.get_xaxis().set_major_formatter(FuncFormatter(lambda x, p: format(int(x), ',')))
        elif a == 'y':
            ax.get_yaxis().set_major_formatter(FuncFormatter(lambda x, p: format(int(x), ',')))


def set_bottom_spine_pos0(ax=None):
    if ax is None:
        ax = plt.gca()
    ax.spines['bottom'].set_position(('data', 0))
